{
  Copyright 1997-2000 by Frank Heckenbach, heckenb@mi.uni-erlangen.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 2.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}

{$ifdef USE_GMP}
{$define _(NAME) NAME##_GMP}
{$else}
{$define _(NAME) NAME}
{$endif}

unit _(Arith);

interface

uses GPC, Commentator_Pas, Tools, Simplicial {$ifdef USE_GMP},Gmp{$endif};

type
  NormalInt = Integer;
  LargeInt = LongInt;

const
  {Make sure that a LargeInt can hold a product of two NormalInt's}
  {If not, compiling will stop here with a "division by zero" error.}
  Assert1=1 div Ord(BitSizeof(LargeInt)>=2*BitSizeof(NormalInt));

type
  {Typ fuer die Werte der Elemente von Vektoren}
  PVectorValue=^TVectorValue;

{$ifdef USE_GMP}

  TVV_GMP=record
  case Boolean of
    False:({Muss zu mpz_t passen}
           Alloc:Integer;
           Val:NormalInt);
    True: (GMPVal:mpz_t)
  end;
  TVectorValue=TVV_GMP;

const
  MaxVal=1 shl (BitSizeOf(NormalInt)-2)-1;

{$else}

  TVectorValue=Integer;

const
  {Maximaler Wert der Vektor-Elemente}
  MaxVV=1 shl (BitSizeOf(TVectorValue)-1)-1;

  {Wurzel davon}
  RootMaxVV=Trunc(Sqrt(MaxVV));

{$endif}

type
  {Typ fuer einen Eintrag in einem "sparse" (duennbesetzten) Vektor}
  PComponent=^TComponent;
  TComponent=record
    Index:TCount;
    Val:TVectorValue
  end;

  {Typ fuer die Eintraege eines sparse Vektors}
  PComponents=^TComponents;
  TComponents=array[0..MaxArray] of TComponent;

  {Typ fuer einen sparse Vektor, bestehend aus den Eintraegen,
   ihrer aktuellen Anzahl und der Anzahl der Eintraege, fuer die Speicher
   reserviert wurde}
  PSparseVector=^TSparseVector;
  TSparseVector=record
    Components:PComponents;
    MaxEl:TVectorValue;
    Count,Size:TCount
  end;
  _(TSV)=TSparseVector;

  {Typ fuer eine sparse Matrix, als Liste aus sparse Vektoren.
   Ausserdem kann hier fuer jede Zeile bei der
   Gauss-Elimination festgehalten werden, in welcher
   Spalte ein Pivotelement auftritt.}
  PPSparseMatrix=^PSparseMatrix;
  PSparseMatrix=^TSparseMatrix;
  TSparseMatrix=record
    Next:PSparseMatrix;
    Row:TSparseVector;
    PivotColumn:TCount
  end;

  {Typ, um die Werte und Anzahl von Torsionen festzuhalten}
  PPTorsionList=^PTorsionList;
  PTorsionList=^TTorsionList;
  TTorsionList=record
    Next:PTorsionList;
    TVal:TVectorValue;
    Count:TCount
  end;

  {Typ fuer eine Tabelle von Homologiegruppen}
  _(THomTable)=array[Dim0] of record
    Betti:TCount;
    Torsion:PTorsionList
  end;

function _(GetNextRow) (const BigFaces,SmallFaces:TFaces;var NextRow:TSparseVector;MinSize:TCount;var RowCount, RowDoneCount:TCount;procedure StatusProc) : Boolean;
function _(QuickCheck) (const BigFaces,SmallFaces:TFaces):TCount;
procedure _(SmithNormalForm) (const BigFaces,SmallFaces:TFaces;MaxRank,Dim:TCount;
                              Mode:Integer; const Coeff:TVectorValue;
                              var PivotCount:TCount;var TorsionList:PTorsionList);
procedure _(WriteStrTorsion) (var s:String;TorsionList:PTorsionList;Flag:Boolean);
procedure _(WriteHomology) (var OutFile:Text;const Homology:_(THomTable); MinDim, MaxDim : TCount);

implementation

{$B-,W-,R-,W+} {No automatic range checking, ranges are checked explicitly here.}

{Anfaengliche Groesse fuer den gerade umgeformten Vektor.
 Dessen Groesse wird bei Bedarf automatisch erhoeht.}
const Size0=$200;

{Zuweisung zwischen maximalen Elementen, nicht noetig in GMP-Version}
inline procedure SetMax(var MaxEl1:TVectorValue;const MaxEl2:TVectorValue);
begin
{$ifndef USE_GMP}
  MaxEl1:=MaxEl2
{$endif}
end;

{Bestimmt betragsmaessig maximales Element eines sparse Vektor}
inline procedure GetMax(var SparseVector:TSparseVector);
var
  i:TCount;
  m:TVectorValue;
begin
{$ifndef USE_GMP}
  m:=0;
  for i:=0 to SparseVector.Count-1 do
    m:=Max(m,Abs(SparseVector.Components^[i].Val));
  SparseVector.MaxEl:=m
{$endif}
end;

{$ifndef USE_GMP}
{Ueberlauf-Fehler}
procedure CoeffError;
begin
  Abort ('Coefficient overflow! Please use the GMP algorithm.')
end;

{Prueft auf Ueberlauf bei der Addition}
inline function CheckAdd(const s1,s2:TVectorValue):TVectorValue;
begin
  if Abs(s1)>MaxVV-Abs(s2) then CoeffError;
  CheckAdd:=s1+s2
end;

{Prueft auf Ueberlauf bei der Multiplikation}
inline function CheckMult(const f1,f2:TVectorValue):TVectorValue;
begin
  if ((Abs(f1)>RootMaxVV) or (Abs(f2)>RootMaxVV)) and_then (Abs(f1)>MaxVV div Abs(f2)) then CoeffError;
  CheckMult:=f1*f2
end;
{$endif}

{Prueft auf moeglichen Ueberlauf bei der Vektor-Addition}
inline procedure CheckMax(var MaxEl:TVectorValue;Vector:PSparseVector;
                          const OtherMax:TVectorValue);
begin
{$ifndef USE_GMP}
  if MaxEl>MaxVV-OtherMax then
    begin
      if Vector<>nil then
        begin
          GetMax(Vector^);
          MaxEl:=Vector^.MaxEl
        end;
      if MaxEl>MaxVV-OtherMax then CoeffError
    end;
  Inc(MaxEl,OtherMax)
{$endif}
end;

{Integer-Grundfunktionen, die fuer Integers fester Groesse und
 Multi-Precision-Integers unterschiedlich implementiert sind}
inline procedure V_Init(var Dest:TVectorValue);
begin
{$ifdef USE_GMP}
  Dest.Alloc:=-1
{$endif}
end;

inline procedure V_Done(var Dest:TVectorValue);
begin
{$ifdef USE_GMP}
  if Dest.Alloc>=0 then mpz_clear(Dest.GMPVal)
{$endif}
end;

inline procedure Set_Int(var Dest:TVectorValue;Src:NormalInt);
begin
{$ifdef USE_GMP}
  V_Done(Dest);
  Dest.Alloc:=-1;
  Dest.Val:=Src
{$else}
  Dest:=Src
{$endif}
end;

{$ifdef USE_GMP}
inline procedure G_Init(var Dest:TVectorValue);
begin
  if Dest.Alloc<0 then mpz_init(Dest.GMPVal)
end;

inline procedure G_CheckInt(var Dest:TVectorValue);
begin
  if (Dest.Alloc>=0) and (mpz_cmp_si(Dest.GMPVal, MaxVal)<=0)
                     and (mpz_cmp_si(Dest.GMPVal,-MaxVal)>=0) then
    Set_Int(Dest,mpz_get_si(Dest.GmpVal))
end;
{$endif}

{Die folgenden Routinen muessen, abhaengig von "$ifdef USE_GMP" entweder mit
 einfachen Zahlen, oder mit einem Typ arbeiten, der zur Laufzeit entweder
 eine einfacher Zahl oder ein GMP-Typ ist. Wichtig: alle Routinen muessen
 auch dann funktionieren, wenn mehrere Parameter identisch sind (oft Dest
 und ein Src-Parameter), d.h. Dest darf nicht beschrieben werden, solange
 noch Werte aus anderen Parametern ausgelesen werden muessen.}

inline procedure Init_V(var Dest:TVectorValue;const Src:TVectorValue);
begin
{$ifdef USE_GMP}
  if Src.Alloc<0
    then
      begin
        Dest.Alloc:=-1;
        Dest.Val:=Src.Val
      end
    else mpz_init_set(Dest.GMPVal,Src.GMPVal)
{$else}
  Dest:=Src
{$endif}
end;

inline procedure Set_V(var Dest:TVectorValue;const Src:TVectorValue);
begin
{$ifdef USE_GMP}
  if Src.Alloc<0
    then Set_Int(Dest,Src.Val)
    else
      if Dest.Alloc<0
        then mpz_init_set(Dest.GMPVal,Src.GMPVal)
        else mpz_set(Dest.GMPVal,Src.GMPVal)
{$else}
  Dest:=Src
{$endif}
end;

inline procedure Abs_V(var Dest:TVectorValue;const Src:TVectorValue);
begin
{$ifdef USE_GMP}
  if Src.Alloc<0
    then Set_Int(Dest,Abs(Src.Val))
    else
      begin
        G_Init(Dest);
        mpz_abs(Dest.GMPVal,Src.GMPVal)
      end
{$else}
  Dest:=Abs(Src)
{$endif}
end;

inline function eq_VI(const Val1:TVectorValue;Val2:NormalInt):Boolean;
begin
{$ifdef USE_GMP}
  if Val1.Alloc<0
    then eq_VI:=Val1.Val=Val2
    else eq_VI:=mpz_cmp_si(Val1.GMPVal,Val2)=0
{$else}
  eq_VI:=Val1=Val2
{$endif}
end;

inline function eq_VV(const Val1,Val2:TVectorValue):Boolean;
begin
{$ifdef USE_GMP}
  if Val1.Alloc<0
    then
      if Val2.Alloc<0
        then eq_VV:=Val1.Val=Val2.Val
        else eq_VV:=mpz_cmp_si(Val2.GMPVal,Val1.Val)=0
    else
      if Val2.Alloc<0
        then eq_VV:=mpz_cmp_si(Val1.GMPVal,Val2.Val)=0
        else eq_VV:=mpz_cmp(Val1.GMPVal,Val2.GMPVal)=0
{$else}
  eq_VV:=Val1=Val2
{$endif}
end;

inline function less_VV(const Val1,Val2:TVectorValue):Boolean;
begin
{$ifdef USE_GMP}
  if Val1.Alloc<0
    then
      if Val2.Alloc<0
        then less_VV:=Val1.Val<Val2.Val
        else less_VV:=mpz_cmp_si(Val2.GMPVal,Val1.Val)>0
    else
      if Val2.Alloc<0
        then less_VV:=mpz_cmp_si(Val1.GMPVal,Val2.Val)<0
        else less_VV:=mpz_cmp(Val1.GMPVal,Val2.GMPVal)<0
{$else}
  less_VV:=Val1<Val2
{$endif}
end;

{$ifdef USE_GMP}
var
  mpz_tmp,mpz_tmp2:mpz_t;
  V_Tmp1,V_Tmp2:TVectorValue;
{$define Declare_V_Tmp}
{$else}
{$define Declare_V_Tmp var V_Tmp1,V_Tmp2:TVectorValue;}
{$endif}

inline function eq_VV_neg(const Val1,Val2:TVectorValue):Boolean;
begin
{$ifdef USE_GMP}
  if Val1.Alloc<0
    then
      if Val2.Alloc<0
        then eq_VV_neg:=Val1.Val=-Val2.Val
        else eq_VV_neg:=mpz_cmp_si(Val2.GMPVal,-Val1.Val)=0
    else
      if Val2.Alloc<0
        then eq_VV_neg:=mpz_cmp_si(Val1.GMPVal,-Val2.Val)=0
        else
          begin
            mpz_neg(mpz_tmp,Val2.GMPVal);
            eq_VV_neg:=mpz_cmp(Val1.GMPVal,mpz_tmp)=0
          end
{$else}
  eq_VV_neg:=Val1=-Val2
{$endif}
end;

{$ifdef USE_GMP}
inline function mpz_IsZero(const Val:mpz_t):Boolean;
begin
  mpz_IsZero:=Val.mp_size=0
end;

inline function mpz_Negative(const Val:mpz_t):Boolean;
begin
  mpz_Negative:=Val.mp_size<0
end;

inline procedure mpz_add_si(var Dest:mpz_t;const Src1:mpz_t;Src2:NormalInt);
begin
  if Src2>=0
    then mpz_add_ui(Dest,Src1,Src2)
    else mpz_sub_ui(Dest,Src1,-Src2)
end;

inline procedure mpz_mul_si(var Dest:mpz_t;const Src1:mpz_t;Src2:NormalInt);
begin
  if Src2>=0
    then mpz_mul_ui(Dest,Src1,Src2)
    else
      begin
        mpz_mul_ui(Dest,Src1,-Src2);
        mpz_neg(Dest,Dest)
      end
end;

inline procedure mpz_tdiv_q_si(var Dest:mpz_t;const Src1:mpz_t;Src2:NormalInt);
begin
  if Src2>=0
    then mpz_tdiv_q_ui(Dest,Src1,Src2)
    else
      begin
        mpz_tdiv_q_ui(Dest,Src1,-Src2);
        mpz_neg(Dest,Dest)
      end
end;

inline procedure mpz_tdiv_qr_si(var DestQ,DestR:mpz_t;const Src1:mpz_t;Src2:NormalInt);
begin
  if Src2>=0
    then mpz_tdiv_qr_ui(DestQ,DestR,Src1,Src2)
    else
      begin
        mpz_tdiv_qr_ui(DestQ,DestR,Src1,-Src2);
        mpz_neg(DestQ,DestQ);
        mpz_neg(DestR,DestR)
      end
end;

inline function mpz_mod_si(const Src1:mpz_t;Src2:NormalInt):MedCard;
begin
  mpz_mod_si:=mpz_fdiv_r_ui(mpz_tmp,Src1,Abs(Src2))
end;
{$endif}

inline function IsZero(const Val:TVectorValue):Boolean;
begin
{$ifdef USE_GMP}
  if Val.Alloc<0
    then IsZero:=Val.Val=0
    else IsZero:=mpz_IsZero(Val.GMPVal)
{$else}
  IsZero:=Val=0
{$endif}
end;

inline function Negative(const Val1:TVectorValue):Boolean;
begin
{$ifdef USE_GMP}
  if Val1.Alloc<0
    then Negative:=Val1.Val<0
    else Negative:=mpz_Negative(Val1.GMPVal)
{$else}
  Negative:=Val1<0
{$endif}
end;

inline procedure Negate(var Dest:TVectorValue;const Src:TVectorValue);
begin
{$ifdef USE_GMP}
  if Src.Alloc<0
    then Set_Int(Dest,-Src.Val)
    else
      begin
        G_Init(Dest);
        mpz_neg(Dest.GMPVal,Src.GMPVal)
      end
{$else}
  Dest:=-Src
{$endif}
end;

inline procedure Add(var Dest:TVectorValue;const Src1,Src2:TVectorValue);
var
  Tmp,Val1,Val2:NormalInt;
  All1,All2:Integer;
begin
{$ifdef USE_GMP}
  if (Src1.Alloc<0) and (Src2.Alloc<0) then
    begin
      Tmp:=Src1.Val+Src2.Val;
      if Abs(Tmp)<=MaxVal
        then Set_Int(Dest,Tmp)
        else
          begin
            G_Init(Dest);
            mpz_set_si(Dest.GMPVal,Tmp)
          end
    end
  else
    begin
      Val1:=Src1.Val;
      Val2:=Src2.Val;
      All1:=Src1.Alloc;
      All2:=Src2.Alloc;
      G_Init(Dest);
      if All1<0 then
        mpz_add_si(Dest.GMPVal,Src2.GMPVal,Val1)
      else if All2<0 then
        mpz_add_si(Dest.GMPVal,Src1.GMPVal,Val2)
      else mpz_add(Dest.GMPVal,Src1.GMPVal,Src2.GMPVal)
    end
{$else}
  Dest:=Src1+Src2
{$endif}
end;

inline procedure Add_Check(var Dest:TVectorValue;const Src1,Src2:TVectorValue);
begin
{$ifdef USE_GMP}
  Add(Dest,Src1,Src2)
{$else}
  Dest:=CheckAdd(Src1,Src2)
{$endif}
end;

inline procedure Sub(var Dest:TVectorValue;const Src1,Src2:TVectorValue);
var
  Tmp,Val1,Val2:NormalInt;
  All1,All2:Integer;
begin
{$ifdef USE_GMP}
  if (Src1.Alloc<0) and (Src2.Alloc<0) then
    begin
      Tmp:=Src1.Val-Src2.Val;
      if Abs(Tmp)<=MaxVal
        then Set_Int(Dest,Tmp)
        else
          begin
            G_Init(Dest);
            mpz_set_si(Dest.GMPVal,Tmp)
          end
    end
  else
    begin
      Val1:=Src1.Val;
      Val2:=Src2.Val;
      All1:=Src1.Alloc;
      All2:=Src2.Alloc;
      G_Init(Dest);
      if All1<0 then
        begin
          mpz_neg(Dest.GMPVal,Src2.GMPVal);
          mpz_add_si(Dest.GMPVal,Dest.GMPVal,Val1)
        end
      else if All2<0 then
        mpz_add_si(Dest.GMPVal,Src1.GMPVal,-Val2)
      else mpz_sub(Dest.GMPVal,Src1.GMPVal,Src2.GMPVal)
    end
{$else}
  Dest:=Src1-Src2
{$endif}
end;

inline procedure Mult(var Dest:TVectorValue;const Src1,Src2:TVectorValue);
var
  Tmp:LargeInt;
  Val1,Val2:NormalInt;
  All1,All2:Integer;
begin
{$ifdef USE_GMP}
  if (Src1.Alloc<0) and (Src2.Alloc<0) then
    begin
      Tmp:=LargeInt(Src1.Val)*LargeInt(Src2.Val);
      if Abs(Tmp)<=MaxVal
        then Set_Int(Dest,Tmp)
        else
          begin
            Val1:=Src1.Val;
            Val2:=Src2.Val;
            G_Init(Dest);
            mpz_set_si(Dest.GMPVal,Val1);
            mpz_mul_si(Dest.GMPVal,Dest.GMPVal,Val2)
          end
    end
  else
    begin
      Val1:=Src1.Val;
      Val2:=Src2.Val;
      All1:=Src1.Alloc;
      All2:=Src2.Alloc;
      G_Init(Dest);
      if All1<0 then
        mpz_mul_si(Dest.GMPVal,Src2.GMPVal,Val1)
      else if All2<0 then
        mpz_mul_si(Dest.GMPVal,Src1.GMPVal,Val2)
      else mpz_mul(Dest.GMPVal,Src1.GMPVal,Src2.GMPVal)
    end
{$else}
  Dest:=Src1*Src2
{$endif}
end;

inline procedure Mult_Check(var Dest:TVectorValue;const Src1,Src2:TVectorValue);
begin
{$ifdef USE_GMP}
  Mult(Dest,Src1,Src2)
{$else}
  Dest:=CheckMult(Src1,Src2)
{$endif}
end;

inline procedure Divide(var Dest:TVectorValue;const Src1,Src2:TVectorValue);
var
  Val1,Val2:NormalInt;
  All1,All2:Integer;
begin
{$ifdef USE_GMP}
  if (Src1.Alloc<0) and (Src2.Alloc<0) then
    Set_Int(Dest,Src1.Val div Src2.Val)
  else
    begin
      Val1:=Src1.Val;
      Val2:=Src2.Val;
      All1:=Src1.Alloc;
      All2:=Src2.Alloc;
      G_Init(Dest);
      if All1<0 then
        begin
          mpz_set_si(Dest.GMPVal,Val1);
          mpz_tdiv_q(Dest.GMPVal,Dest.GMPVal,Src2.GMPVal)
        end
      else if All2<0 then
        mpz_tdiv_q_si(Dest.GMPVal,Src1.GMPVal,Val2)
      else mpz_tdiv_q(Dest.GMPVal,Src1.GMPVal,Src2.GMPVal);
      G_CheckInt(Dest)
    end
{$else}
  Dest:=Src1 div Src2
{$endif}
end;

inline procedure DivExact(var Dest:TVectorValue;const Src1,Src2:TVectorValue);
var
  Val1,Val2:NormalInt;
  All1,All2:Integer;
begin
{$ifdef USE_GMP}
  if (Src1.Alloc<0) and (Src2.Alloc<0) then
    Set_Int(Dest,Src1.Val div Src2.Val)
  else
    begin
      Val1:=Src1.Val;
      Val2:=Src2.Val;
      All1:=Src1.Alloc;
      All2:=Src2.Alloc;
      G_Init(Dest);
      if All1<0 then
        begin
          mpz_set_si(Dest.GMPVal,Val1);
          mpz_tdiv_q(Dest.GMPVal,Dest.GMPVal,Src2.GMPVal)
        end
      else if All2<0 then
        mpz_tdiv_q_si(Dest.GMPVal,Src1.GMPVal,Val2)
      else mpz_divexact(Dest.GMPVal,Src1.GMPVal,Src2.GMPVal);
      G_CheckInt(Dest)
    end
{$else}
  Dest:=Src1 div Src2
{$endif}
end;

inline function Divisible(const Src1,Src2:TVectorValue):Boolean;
begin
{$ifdef USE_GMP}
  if (Src1.Alloc<0) and (Src2.Alloc<0) then
    Divisible:=Src1.Val mod Src2.Val=0
  else
    begin
      if Src1.Alloc<0 then
        begin
          mpz_set_si(mpz_tmp,Src1.Val);
          mpz_mod(mpz_tmp,mpz_tmp,Src2.GMPVal);
          Divisible:=mpz_IsZero(mpz_tmp)
        end
      else if Src2.Alloc<0 then
        Divisible:=mpz_mod_si(Src1.GMPVal,Src2.Val)=0
      else
        begin
          mpz_mod(mpz_tmp,Src1.GMPVal,Src2.GMPVal);
          Divisible:=mpz_IsZero(mpz_tmp)
        end
    end
{$else}
  Divisible:=Src1 mod Src2=0
{$endif}
end;

inline function Divisible_F(var Dest:TVectorValue;const Src1,Src2:TVectorValue):Boolean;
var
  Val1,Val2:NormalInt;
  All1,All2:Integer;
begin
{$ifdef USE_GMP}
  if (Src1.Alloc<0) and (Src2.Alloc<0) then
    begin
      if Src1.Val mod Src2.Val=0
        then
          begin
            Set_Int(Dest,Src1.Val div Src2.Val);
            Divisible_F:=True
          end
        else Divisible_F:=False
    end
  else
    begin
      Val1:=Src1.Val;
      Val2:=Src2.Val;
      All1:=Src1.Alloc;
      All2:=Src2.Alloc;
      G_Init(Dest);
      if All1<0 then
        begin
          mpz_set_si(Dest.GMPVal,Val1);
          mpz_tdiv_qr(Dest.GMPVal,mpz_tmp,Dest.GMPVal,Src2.GMPVal)
        end
      else if All2<0 then
        mpz_tdiv_qr_si(Dest.GMPVal,mpz_tmp,Src1.GMPVal,Val2)
      else
        mpz_tdiv_qr(Dest.GMPVal,mpz_tmp,Src1.GMPVal,Src2.GMPVal);
      Divisible_F:=mpz_IsZero(mpz_tmp);
      G_CheckInt(Dest)
    end
{$else}
  if Src1 mod Src2=0
    then
      begin
        Dest:=Src1 div Src2;
        Divisible_F:=True
      end
    else Divisible_F:=False
{$endif}
end;

inline procedure ModReduce(var v:TVectorValue;const Modulus:TVectorValue);
begin
{$ifdef USE_GMP}
  if v.Alloc<0 then
    if (Modulus.Alloc<0) and (Abs(v.Val)>=Abs(Modulus.Val)) then v.Val:=v.Val mod Modulus.Val else
  else
    begin
      if Modulus.Alloc<0 then
        begin
          mpz_set_si(mpz_tmp,Modulus.Val);
          mpz_mod(v.GMPVal,v.GMPVal,mpz_tmp)
        end
      else
        if mpz_sizeinbase (v.GMPVal,2) >= mpz_sizeinbase (Modulus.GMPVal,2) then
          mpz_mod(v.GMPVal,v.GMPVal,Modulus.GMPVal);
      G_CheckInt(v)
    end
{$else}
  if Abs(v)>=Abs(Modulus) then v:=v mod Modulus
{$endif}
end;

{$ifndef USE_GMP}
{Erweiterter euklidischer Algorithmus
 Der Rueckgabewert ist der (positive) ggT von a und b, und es gilt
 /c d\ /a\ = /ggT\ und det /c d\ = +/-1
 \e f/ \b/   \ 0 /         \e f/        }
inline procedure Int_ggTExt(var Dest:TVectorValue;const a0,b0:TVectorValue;var c,d,e,f:TVectorValue);
var a,b,g,h:TVectorValue;
begin
  a:=a0;
  b:=b0;
  c:=1;
  d:=0;
  e:=0;
  f:=1;
  if a<0 then
    begin
      a:=-a;
      c:=-1
    end;
  if b<0 then
    begin
      b:=-b;
      f:=-1
    end;
  while b<>0 do
    begin
      g:=a div b;
      h:=CheckAdd(c,-CheckMult(g,e));
      c:=e;
      e:=h;
      h:=CheckAdd(d,-CheckMult(g,f));
      d:=f;
      f:=h;
      h:=CheckAdd(a,-CheckMult(g,b));
      a:=b;
      b:=h
    end;
  Dest:=a
end;
{$endif}

inline procedure ggTExt(var Dest:TVectorValue;const Src1,Src2:TVectorValue;var c,d,e,f:TVectorValue);
begin
{$ifdef USE_GMP}
  if Src1.Alloc<0
    then mpz_set_si(mpz_tmp,Src1.Val)
    else mpz_set(mpz_tmp,Src1.GMPVal);
  if Src2.Alloc<0
    then mpz_set_si(mpz_tmp2,Src2.Val)
    else mpz_set(mpz_tmp2,Src2.GMPVal);
  G_Init(Dest);
  G_Init(c);
  G_Init(d);
  G_Init(e);
  G_Init(f);
  mpz_gcdext(Dest.GMPVal,c.GMPVal,d.GMPVal,mpz_tmp,mpz_tmp2);
  mpz_divexact(e.GMPVal,mpz_tmp2,Dest.GMPVal);
  mpz_divexact(f.GMPVal,mpz_tmp,Dest.GMPVal);
  mpz_neg(f.GMPVal,f.GMPVal);
  G_CheckInt(Dest);
  G_CheckInt(c);
  G_CheckInt(d);
  G_CheckInt(e);
  G_CheckInt(f);
{$else}
  Int_ggTExt(Dest,Src1,Src2,c,d,e,f)
{$endif}
end;

inline procedure ggT(var Dest:TVectorValue;const Src1,Src2:TVectorValue);
{$ifdef USE_GMP}
begin
  if Src1.Alloc<0
    then mpz_set_si(mpz_tmp,Src1.Val)
    else mpz_set(mpz_tmp,Src1.GMPVal);
  if Src2.Alloc<0
    then mpz_set_si(mpz_tmp2,Src2.Val)
    else mpz_set(mpz_tmp2,Src2.GMPVal);
  G_Init(Dest);
  mpz_gcd(Dest.GMPVal,mpz_tmp,mpz_tmp2);
  G_CheckInt(Dest)
{$else}
var d1,d2,d3,d4:TVectorValue;
begin
  Int_ggTExt(Dest,Src1,Src2,d1,d2,d3,d4)
{$endif}
end;

procedure V_Write(var OutFile:Text;const Val:TVectorValue);
var Str:CString;
begin
{$ifdef USE_GMP}
  if Val.Alloc<0
    then Write(OutFile,Val.Val)
    else
      begin
        Str:=mpz_get_str(nil,10,Val.GMPVal);
        {$X+}Write(OutFile,Str);{$X-}
        Dispose(Str)
      end
{$else}
  Write(OutFile,Val)
{$endif}
end;

procedure V_WriteStr(var s:String;const Val:TVectorValue);
var Str:CString;
begin
{$ifdef USE_GMP}
  if Val.Alloc<0
    then WriteStr(s,Val.Val)
    else
      begin
        Str:=mpz_get_str(nil,10,Val.GMPVal);
        s:=CString2String(Str);
        Dispose(Str)
      end
{$else}
  WriteStr(s,Val)
{$endif}
end;

procedure V_WriteStr2(var s:String;const Val:TVectorValue);
var
  Str:CString;
  Digits:SizeType;
begin
{$ifdef USE_GMP}
  if Val.Alloc<0 then
    WriteStr(s,Val.Val)
  else
    begin
      Digits := mpz_sizeinbase (Val.GMPVal, 10);
      if Digits <= 24 then
        begin
          Str:=mpz_get_str(nil,10,Val.GMPVal);
          s:=CString2String(Str);
          Dispose(Str)
        end
      else
        WriteStr (s, '(', Digits, ' digits)')
    end
{$else}
  WriteStr(s,Val)
{$endif}
end;

{Legt einen neuen sparse Vektor an}
inline procedure NewSparseVector(var SparseVector:TSparseVector;InitStart:TCount);
var i:TCount;
begin
  if SparseVector.Size>=MaxArray then Abort ('Vector too long!');
  GetMem(SparseVector.Components,SparseVector.Size*SizeOf(SparseVector.Components^[0]));
  {$ifdef USE_GMP}
  for i:=InitStart to SparseVector.Size-1 do V_Init(SparseVector.Components^[i].Val)
  {$endif}
end;

{Erzeugt eine Kopie eines sparse Vektors}
inline procedure CopySparseVector(Source:TSparseVector;var Dest:TSparseVector);
var i:TCount;
begin
  Dest.Count:=Source.Count;
  Dest.Size:=Dest.Count;
  SetMax(Dest.MaxEl,Source.MaxEl);
  NewSparseVector(Dest,0);
  for i:=0 to Dest.Count-1 do
    begin
      Dest.Components^[i].Index:=Source.Components^[i].Index;
      Set_V(Dest.Components^[i].Val,Source.Components^[i].Val)
    end
end;

{Vergroessert einen sparse Vector}
inline procedure ResizeSparseVector(var Vector:TSparseVector;NewSize:TCount);
var
  VectorT:TSparseVector;
  i:TCount;
begin
  VectorT.Count:=Vector.Count;
  VectorT.Size:=Max(NewSize,Vector.Size);
  SetMax(VectorT.MaxEl,Vector.MaxEl);
  NewSparseVector(VectorT,Vector.Size);
  for i:=0 to Vector.Size-1 do VectorT.Components^[i]:=Vector.Components^[i];
  Dispose(Vector.Components);
  Vector:=VectorT
end;

{Entfernt einen sparse Vektor}
inline procedure DisposeSparseVector(SparseVector:TSparseVector;DoneStart:TCount);
var i:TCount;
begin
  {$ifdef USE_GMP}
  for i:=DoneStart to SparseVector.Size-1 do V_Done(SparseVector.Components^[i].Val);
  {$endif}
  Dispose(SparseVector.Components)
end;

{Entfernt eine Matrix aus sparse Vektoren}
procedure DisposeMatrix(Matrix:PSparseMatrix);
var Temp:PSparseMatrix;
begin
  while Matrix<>nil do
    begin
      Temp:=Matrix;
      Matrix:=Matrix^.Next;
      DisposeSparseVector(Temp^.Row,0);
      Dispose(Temp)
    end
end;

{Multipliziert einen sparse Vektor mit -1}
inline procedure SparseNeg(SparseVector:TSparseVector);
var c:TCount;
begin
  for c:=0 to SparseVector.Count-1 do
    with SparseVector.Components^[c] do Negate(Val,Val)
end;

{Dividiert einen sparse Vektor durch eine ganze Zahl, wenn bekannt ist, dass
 die Division aufgeht}
inline procedure SparseDivExact(SparseVector:TSparseVector;const Divisor:TVectorValue);
var c:TCount;
begin
{$ifndef USE_GMP}
  SparseVector.MaxEl:=SparseVector.MaxEl div Divisor;
{$endif}
  for c:=0 to SparseVector.Count-1 do
    with SparseVector.Components^[c] do DivExact(Val,Val,Divisor)
end;

{$X+} {Extended syntax for pointer arithmetic}
{Wichtig: Komponenten, die bei der Umformung zu 0 werden, duerfen im
 Ergebnisvektor nicht gespeichert werden.}
{$define Store(Idx,NoTest)
  if NoTest or not IsZero(V_Tmp1) then
    begin
      vr^.Index:=Idx;
      Set_V(vr^.Val,V_Tmp1);
      Inc(vr)
    end
}

{Makro fuer die verschiedenen Zeilenumformungen. Op1 und Op2 sind die auf
 Elemente von SparseVector1 bzw. SparseVector2 anzuwendenden Operationen,
 die nicht miteinander verknuepft werden, und Op12 ist die Operation, die
 Elemente beider Vektoren verknuepft. Die Ergebnisse der Operationen
 muessen jeweils in V_Tmp1 abgelegt werden.}
{$define SparseAddGeneric(NoTest1,NoTest2,Op1,Op2,Op12)
var
  c1,c2:TCount;
  v1,v2,vr:PComponent;
  Declare_V_Tmp
begin
  v1:=@SparseVector1.Components^[0];
  v2:=@SparseVector2.Components^[0];
  vr:=@Result.Components^[0];
  c1:=SparseVector1.Count;
  c2:=SparseVector2.Count;
  if (c1>0) and (c2>0) then
    repeat
      if v1^.Index=v2^.Index then
        begin
          Op12;
          Store(v1^.Index,False);
          Inc(v1);
          Inc(v2);
          Dec(c1);
          Dec(c2);
          if (c1=0) or (c2=0) then Break
        end
      else if v1^.Index<v2^.Index then
        begin
          Op1;
          Store(v1^.Index,NoTest1);
          Inc(v1);
          Dec(c1);
          if c1=0 then Break
        end
      else
        begin
          Op2;
          Store(v2^.Index,NoTest2);
          Inc(v2);
          Dec(c2);
          if c2=0 then Break
        end
    until False;
  while c1>0 do
    begin
      Op1;
      Store(v1^.Index,NoTest1);
      Inc(v1);
      Dec(c1)
    end;
  while c2>0 do
    begin
      Op2;
      Store(v2^.Index,NoTest2);
      Inc(v2);
      Dec(c2)
    end;
  Result.Count:=vr-@Result.Components^[0]
end;
}

{Addiert einen sparse Vektor zu einem anderen sparse Vektor}
procedure SparseAdd(const SparseVector1,SparseVector2:TSparseVector;
                    var Result:TSparseVector);
SparseAddGeneric(True,True,
  Set_V(V_Tmp1,v1^.Val),
  Set_V(V_Tmp1,v2^.Val),
  Add(V_Tmp1,v1^.Val,v2^.Val))

{Addiert einen sparse Vektor zu einem anderen sparse Vektor modulo Modulus}
procedure SparseAddMod(const SparseVector1,SparseVector2:TSparseVector;
                       var Result:TSparseVector;
                       const Modulus:TVectorValue);
SparseAddGeneric(False,False,
  Set_V(V_Tmp1,v1^.Val),
  Set_V(V_Tmp1,v2^.Val),
  Add(V_Tmp1,v1^.Val,v2^.Val);
  ModReduce(V_Tmp1,Modulus))

{Subtrahiert einen sparse Vektor von einem anderen sparse Vektor}
procedure SparseSub(const SparseVector1,SparseVector2:TSparseVector;
                    var Result:TSparseVector);
SparseAddGeneric(True,True,
  Set_V(V_Tmp1,v1^.Val),
  Negate(V_Tmp1,v2^.Val),
  Sub(V_Tmp1,v1^.Val,v2^.Val))

{Addiert das ganzzahlige Vielfache eines sparse Vektors zu einem
 anderen sparse Vektor}
procedure SparseAddScaled(const SparseVector1,SparseVector2:TSparseVector;
                          const ScaleFactor2:TVectorValue;
                          var Result:TSparseVector);
SparseAddGeneric(True,True,
  Set_V(V_Tmp1,v1^.Val),
  Mult(V_Tmp1,ScaleFactor2,v2^.Val),
  Mult(V_Tmp2,ScaleFactor2,v2^.Val);
  Add(V_Tmp1,v1^.Val,V_Tmp2))

{Addiert das ganzzahlige Vielfache eines sparse Vektors zu einem
 anderen sparse Vektor modulo Modulus}
procedure SparseAddScaledMod(const SparseVector1,SparseVector2:TSparseVector;
                             const ScaleFactor2:TVectorValue;
                             var Result:TSparseVector;
                             const Modulus:TVectorValue);
SparseAddGeneric(False,False,
  Set_V(V_Tmp1,v1^.Val),
  Mult(V_Tmp1,ScaleFactor2,v2^.Val);
  ModReduce(V_Tmp1,Modulus),
  Mult(V_Tmp2,ScaleFactor2,v2^.Val);
  Add(V_Tmp1,v1^.Val,V_Tmp2);
  ModReduce(V_Tmp1,Modulus))

{Addiert das ganzzahlige Vielfache eines sparse Vektors zu dem
 Vielfachen eines anderen sparse Vektors}
procedure SparseAddScaledBoth(SparseVector1,SparseVector2:TSparseVector;
                              const ScaleFactor1,ScaleFactor2:TVectorValue;
                              var Result:TSparseVector);
SparseAddGeneric(False,False,
  Mult(V_Tmp1,ScaleFactor1,v1^.Val),
  Mult(V_Tmp1,ScaleFactor2,v2^.Val),
  Mult(V_Tmp1,ScaleFactor1,v1^.Val);
  Mult(V_Tmp2,ScaleFactor2,v2^.Val);
  Add(V_Tmp1,V_Tmp1,V_Tmp2))
{$X-}

{Addiert das ganzzahlige Vielfache einer Spalte einer Matrix zu einer
 anderen Spalte der Matrix modulo Modulus}
procedure AddColumnScaledMod(Matrix:PSparseMatrix;
                             SourceColumn,DestColumn:TCount;
                             const ScaleFactor:TVectorValue;
                             const Modulus:TVectorValue);
var
  i,j,k:TCount;
  V_Temp:TVectorValue;
begin
  V_Init(V_Temp);
  while Matrix<>nil do
    with Matrix^.Row do
      begin
        i:=0;
        while (i<Count) and (Components^[i].Index<SourceColumn) do Inc(i);
        if (i<Count) and (Components^[i].Index=SourceColumn) then
          begin
            if DestColumn<SourceColumn then j:=0 else j:=i;
            while (j<Count) and (Components^[j].Index<DestColumn) do Inc(j);
            if (j<Count) and (Components^[j].Index=DestColumn)
              then
                with Components^[j] do
                  begin
                    Mult_Check(V_Temp,ScaleFactor,Components^[i].Val);
                    Add_Check(Val,Val,V_Temp);
                    ModReduce(Val,Modulus);
                    if IsZero(Val) then
                      begin
                        V_Done(Val);
                        Dec(Count);
                        for k:=j to Count-1 do Components^[k]:=Components^[k+1];
                        V_Init(Components^[Count].Val)
                      end
                  end
              else
                begin
                  Mult_Check(V_Temp,ScaleFactor,Components^[i].Val);
                  ModReduce(V_Temp,Modulus);
                  if not IsZero(V_Temp) then
                    begin
                      if Size<=Count then ResizeSparseVector(Matrix^.Row,Count+4);
                      V_Done(Components^[Count].Val);
                      for k:=Count-1 downto j do Components^[k+1]:=Components^[k];
                      Inc(Count);
                      Init_V(Components^[j].Val,V_Temp);
                      Components^[j].Index:=DestColumn
                    end
                end
          end;
        Matrix:=Matrix^.Next
      end;
  V_Done(V_Temp)
end;

{Reduziert den Vektor mit allen Zeilen der Matrix, deren Pivotelement 1 ist}
procedure SparseReduce(var Vector:TSparseVector;RemainingMatrix:PSparseMatrix);
var
  Vector2,VectorT:TSparseVector;
  i,NewSize:TCount;
  MaxEl,V_Temp:TVectorValue;
  Val:PVectorValue;
begin
  V_Init(V_Temp);
  SetMax(MaxEl,Vector.MaxEl);
  Vector2.Size:=Max(Vector.Size,Size0);
  NewSparseVector(Vector2,0);
  while RemainingMatrix<>nil do
    begin
      if eq_VI(RemainingMatrix^.Row.Components^[0].Val,1) then
        begin
          i:=0;
          while (i<Vector.Count) and
            (Vector.Components^[i].Index<RemainingMatrix^.PivotColumn) do Inc(i);
          if (i<Vector.Count) and
            (Vector.Components^[i].Index=RemainingMatrix^.PivotColumn) then
            begin
              if Vector2.Size<Vector.Count+RemainingMatrix^.Row.Count then
                begin
                  NewSize:=Vector2.Size;
                  repeat
                    NewSize:=2*NewSize
                  until NewSize>=Vector.Count+RemainingMatrix^.Row.Count;
                  Vector2.Count:=0;
                  ResizeSparseVector(Vector2,NewSize)
                end;
              Val:=@Vector.Components^[i].Val;
              if eq_VI(Val^,1) then
                begin
                  CheckMax(MaxEl,@Vector,RemainingMatrix^.Row.MaxEl);
                  SparseSub(Vector,RemainingMatrix^.Row,Vector2)
                end
              else if eq_VI(Val^,-1) then
                begin
                  CheckMax(MaxEl,@Vector,RemainingMatrix^.Row.MaxEl);
                  SparseAdd(Vector,RemainingMatrix^.Row,Vector2)
                end
              else
                begin
                  {$ifndef USE_GMP}
                  CheckMax(MaxEl,@Vector,CheckMult(RemainingMatrix^.Row.MaxEl,Abs(Val^)));
                  {$endif}
                  Negate(V_Temp,Val^);
                  SparseAddScaled(Vector,RemainingMatrix^.Row,V_Temp,Vector2)
                end;
              VectorT:=Vector;
              Vector:=Vector2;
              Vector2:=VectorT
            end
        end;
      RemainingMatrix:=RemainingMatrix^.Next
    end;
  GetMax(Vector);
  DisposeSparseVector(Vector2,0);
  V_Done(V_Temp)
end;

{Zaehlt die Anzahl der offensichtlich linear unabhaengigen Matrixzeilen,
 um ggf. die Gauss-Elimination zu sparen, falls im d_p = ker d_(p-1)}
function _(QuickCheck) (const BigFaces,SmallFaces:TFaces):TCount;
var
  LastCol,CurrentCol,b,c:TCount;
  Temp:PSparseIntVector;
begin
  c:=0;
  for b:=FacesCount(BigFaces)-1 downto 0 do
    begin
      Temp:=GetOneRow(BigFaces,SmallFaces,b);
      if Temp^.Count<>0 then
        begin
          CurrentCol:=Temp^.Elements[0].Index;
          if (c=0) or (CurrentCol>LastCol)
            then
              begin
                LastCol:=CurrentCol;
                Inc(c)
              end
            else
              if CurrentCol<LastCol then return c { Hier wird der Test abgebrochen, weil er mit diesen einfachen Mitteln nicht fortgesetzt werden kann. }
        end;
      Dispose(Temp)
    end;
  _(QuickCheck) :=c
end;

procedure GetTorsion(Matrix:PSparseMatrix;Rows:TCount;TorsionList:PPTorsionList;Mode:Integer; const Coeff:TVectorValue);
var
  TorsionCount,min_Col,i,j,c,c2:TCount;
  v:PVectorValue;
  min_Row,Temp:PSparseMatrix;
  PTemp:PPSparseMatrix;
  Vector2:TSparseVector;
  v_abs,TorsionVal,det,m_ggT,m_min,m_min_abs,MaxEl,Factor:TVectorValue;
  Msg:TString;
  MsgLevel:Integer;

  procedure StoreTorsion;
  begin
    if not eq_VI(TorsionVal,1) and (TorsionCount<>0) then
      begin
        New(TorsionList^);
        Init_V(TorsionList^^.TVal,TorsionVal);
        TorsionList^^.Count:=TorsionCount;
        TorsionList:=@TorsionList^^.Next;
        TorsionList^:=nil
      end
  end;

  {Row:=Row-(Row.Components^[Index].Val div m_min)*min_Row^.Row}
  inline procedure AddRow(Index:TCount;var Row:TSparseVector);
  var MaxEl:TVectorValue;
  begin
    Vector2.Size:=min_Row^.Row.Count+Row.Count;
    NewSparseVector(Vector2,0);
    Divide(Factor,Row.Components^[Index].Val,m_min);
    Negate(Factor,Factor);
    {$ifndef USE_GMP}
    SetMax(MaxEl,Row.MaxEl);
    CheckMax(MaxEl,nil,CheckMult(min_Row^.Row.MaxEl,Abs(Factor)));
    {$endif}
    SparseAddScaledMod(Row,min_Row^.Row,Factor,Vector2,det);
    GetMax(Vector2);
    DisposeSparseVector(Row,0);
    Row:=Vector2
  end;

begin
  StartActivity (Commentator, 'torsion', 'Computing torsion coefficients', LvlNormal, InternalDescription);
  c:=0;
  c2:=0;
  V_Init(det);
  V_Init(m_min);
  V_Init(m_min_abs);
  V_Init(m_ggT);
  V_Init(Factor);
  V_Init(v_abs);
  V_Init(TorsionVal);
  Set_Int(det,1);
  Temp := Matrix;
  while Temp <> nil do
    with Temp^ do
      begin
        if (Row.Count=0) or (Row.Components^[0].Index<>PivotColumn) then Abort ('Internal error #1 in GetTorsion.');
        Mult_Check(det,det,Row.Components^[0].Val);
        Temp := Temp^.Next
      end;
  if IsPrinted (Commentator, LvlUnimp, InternalDescription) <> 0 then
    begin
      V_WriteStr2(Msg,det);
      ActivityReport (Commentator, 'det(max. minor): ' + Msg, LvlUnimp, InternalDescription)
    end;
  Add_Check(det,det,det);
  Set_Int(TorsionVal,1);
  TorsionCount:=0;
  repeat
    {Bestimme ggT und Minimum der Matrixelemente}
    with Matrix^.Row.Components^[0] do
      begin
        Set_V(m_min,Val);
        Abs_V(m_min_abs,m_min);
        Set_V(m_ggT,m_min_abs);
        min_Col:=Index;
        min_Row:=Matrix;
        if IsZero(m_min) then Abort ('Internal error #2 in GetTorsion.')
      end;
    Temp:=Matrix;
    while Temp<>nil do
      begin
        for i:=0 to Temp^.Row.Count-1 do
          with Temp^.Row.Components^[i] do
            begin
              v:=@Val;
              Abs_V(v_abs,v^);
              if less_VV(v_abs,m_min_abs) then
                begin
                  Set_V(m_min,v^);
                  Set_V(m_min_abs,v_abs);
                  min_Col:=Index;
                  min_Row:=Temp;
                  if IsZero(m_min) then Abort ('Internal error #3 in GetTorsion.')
                end;
              if not Divisible(v^,m_ggT) then ggT(m_ggT,m_ggT,v^);
              if eq_VI(v_abs,1) then Break
            end;
        if eq_VI(m_min_abs,1) then Break;
        Temp:=Temp^.Next
      end;
    {Wenn der ggT groesser als 1 ist, wird dies festgehalten, und alle
     Matrixelemente dadurch dividiert}
    if not eq_VI(m_ggT,1) then
      begin
        StoreTorsion;
        Mult(TorsionVal,TorsionVal,m_ggT);
        if (Mode = hmNTorsionRank) and Divisible(TorsionVal,Coeff) then
          begin
            DisposeMatrix (Matrix);
            TorsionCount := Rows - c;
            Break
          end;
        TorsionCount:=0;
        DivExact(det,det,m_ggT);
        DivExact(m_min,m_min,m_ggT);
        Abs_V(m_min_abs,m_min);
        Temp:=Matrix;
        while Temp<>nil do
          begin
            SparseDivExact(Temp^.Row,m_ggT);
            Temp:=Temp^.Next
          end
      end;
    Inc(c2);
    if c2 mod 10 = 0
      then MsgLevel := LvlUnimp
      else MsgLevel := LvlBlabla;
    if IsPrinted (Commentator, MsgLevel, InternalDescription) <> 0 then
      begin
        V_WriteStr2(Msg,m_min_abs);
        ActivityReport (Commentator, 'Current min: ' + Msg, MsgLevel, InternalDescription)
      end;
    if not eq_VI(m_min_abs,1)
      then
        begin
          {Suche Element in derselben Spalte wie m_min, das kein Vielfaches von m_min ist}
          Temp:=Matrix;
          while Temp<>nil do
            begin
              if Temp<>min_Row then
                begin
                  i:=0;
                  while (i<Temp^.Row.Count) and
                        (Temp^.Row.Components^[i].Index<min_Col) do Inc(i);
                  if i<Temp^.Row.Count then
                    with Temp^.Row.Components^[i] do
                      if (Index=min_Col) and not Divisible(Val,m_min) then Break
                end;
              Temp:=Temp^.Next
            end;
          if Temp<>nil then
            begin
              AddRow(i,Temp^.Row);
              Continue
            end;
          {Suche Element in derselben Zeile wie m_min, das kein Vielfaches von m_min ist}
          i:=0;
          while (i<min_Row^.Row.Count) and
            Divisible(min_Row^.Row.Components^[i].Val,m_min) do Inc(i);
          if i<min_Row^.Row.Count then
            begin
              Divide(Factor,min_Row^.Row.Components^[i].Val,m_min);
              Negate(Factor,Factor);
              AddColumnScaledMod(Matrix,min_Col,min_Row^.Row.Components^[i].Index,Factor,det);
              Continue
            end;
          {Suche Element in einer anderen Zeile und Spalte als m_min, das kein
           Vielfaches von m_min ist}
          Temp:=Matrix;
          while Temp<>nil do
            begin
              i:=0;
              while (i<Temp^.Row.Count) and
                    Divisible(Temp^.Row.Components^[i].Val,m_min) do Inc(i);
              if i<Temp^.Row.Count then Break;
              Temp:=Temp^.Next
            end;
          if Temp=nil then Abort ('Internal error #4 in GetTorsion.');
          j:=0;
          while (j<Temp^.Row.Count) and
                (Temp^.Row.Components^[j].Index<min_Col) do Inc(j);
          if (j<Temp^.Row.Count) and
             (Temp^.Row.Components^[j].Index=min_Col) then AddRow(j,Temp^.Row);
          Vector2.Size:=min_Row^.Row.Count+Temp^.Row.Count;
          NewSparseVector(Vector2,0);
          SetMax(MaxEl,min_Row^.Row.MaxEl);
          CheckMax(MaxEl,nil,Temp^.Row.MaxEl);
          SparseAddMod(min_Row^.Row,Temp^.Row,Vector2,det);
          GetMax(Vector2);
          DisposeSparseVector(min_Row^.Row,0);
          min_Row^.Row:=Vector2
          {Jetzt gibt es ein Element in derselben Zeile wie m_min, das kein
           Vielfaches von m_min ist, und im naechsten Durchgang kann m_min
           verringert werden}
        end
      else
        begin
          {Jetzt ist m_min=+/-1, d.h. teilt alle Matrixelemente, und eine
           Zeile und eine Spalte koennen entfernt werden}
          PTemp:=@Matrix;
          while PTemp^<>nil do
            if PTemp^=min_Row
              then PTemp^:=PTemp^^.Next
              else
                begin
                  i:=0;
                  while (i<PTemp^^.Row.Count) and
                        (PTemp^^.Row.Components^[i].Index<min_Col) do Inc(i);
                  if (i<PTemp^^.Row.Count) and
                     (PTemp^^.Row.Components^[i].Index=min_Col) then
                    AddRow(i,PTemp^^.Row);
                  PTemp:=@PTemp^^.Next
                end;
          DisposeSparseVector(min_Row^.Row,0);
          Dispose(min_Row);
          Inc(TorsionCount);
          Inc(c);
          c2 := 0;
          ProgressReport (Commentator, 'rows', LvlUnimp, c, Rows)
        end
  until Matrix=nil;
  StoreTorsion;
  V_Done(det);
  V_Done(m_min);
  V_Done(m_min_abs);
  V_Done(m_ggT);
  V_Done(Factor);
  V_Done(v_abs);
  V_Done(TorsionVal);
  StopActivity (Commentator, 'Torsion coefficients computed', LvlNormal, InternalDescription)
end;

function _(GetNextRow) (const BigFaces,SmallFaces:TFaces;var NextRow:TSparseVector;MinSize:TCount;var RowCount, RowDoneCount:TCount;procedure StatusProc) : Boolean;
var
  Temp:PSparseIntVector;
  c:TCount;
begin
  repeat
    if RowCount>=FacesCount(BigFaces) then
      begin
        _(GetNextRow):=False;
        Exit
      end;
    Temp:=GetOneRow(BigFaces,SmallFaces,RowCount);
    Inc(RowCount);
    c:=Temp^.Count;
    if c=0 then
      begin
        Inc(RowDoneCount);
        if (MsgStep<>0) and (RowDoneCount mod MsgStep=0) then StatusProc;
        Dispose(Temp)
      end
  until c<>0;
  NextRow.Count:=Temp^.Count;
  NextRow.Size:=Max(NextRow.Count,MinSize);
  NewSparseVector(NextRow,0);
  for c:=0 to Temp^.Count-1 do
    begin
      NextRow.Components^[c].Index:=Temp^.Elements[c].Index;
      Set_Int(NextRow.Components^[c].Val,Temp^.Elements[c].Val)
    end;
  GetMax(NextRow);
  Dispose(Temp);
  _(GetNextRow):=True
end;

{Bestimmt die Smith-Normalform einer sparse Matrix}
procedure _(SmithNormalForm) (const BigFaces,SmallFaces:TFaces;MaxRank,Dim:TCount;
                              Mode:Integer; const Coeff:TVectorValue;
                              var PivotCount:TCount;var TorsionList:PTorsionList);
var
  PivotN1Count,i,LastCol,RowCount,RowDoneCount:TCount;
  ReducedMatrixStartRow:PPSparseMatrix;
  Temp,ReducedMatrix,DeferredVectors,TorsionMatrix:PSparseMatrix;
  Vector,Vector2:TSparseVector;
  pg,vg,p0,v0,Factor,V_Temp:TVectorValue;
  Msg:TString;

  procedure RowStatus;
  begin
    ProgressReport (Commentator, 'rows', LvlUnimp, RowDoneCount, FacesCount (BigFaces));
    WriteStr (Msg, 'current torsion-free rank: ', PivotCount - PivotN1Count,
                   ', current rank: ', PivotCount,
                   ', max. rank: ', MaxRank);
    ActivityReport (Commentator, Msg, LvlUnimp, InternalDescription)
  end;

  procedure DoOneRow(Vector:TSparseVector;Defer:Boolean);
  var
    i,Index1,j,NewSize:TCount;
    NewRow:PSparseMatrix;
    ReducedMatrixRow:PPSparseMatrix;
    VectorT:TSparseVector;
    Val,PivotValue:PVectorValue;
    MaxEl:TVectorValue;
  begin
    if (MsgStep<>0) and (RowDoneCount mod MsgStep=0) then RowStatus;
    SetMax(MaxEl,Vector.MaxEl);
    Index1:=Vector.Components^[0].Index;
    if Index1<LastCol then ReducedMatrixStartRow:=@ReducedMatrix;
    while (ReducedMatrixStartRow^<>nil) and
      (Index1>ReducedMatrixStartRow^^.PivotColumn) do
      ReducedMatrixStartRow:=@ReducedMatrixStartRow^^.Next;
    LastCol:=Index1;
    ReducedMatrixRow:=ReducedMatrixStartRow;
    while ReducedMatrixRow^<>nil do
      if Index1>ReducedMatrixRow^^.PivotColumn
        then ReducedMatrixRow:=@ReducedMatrixRow^^.Next
      else if Index1<ReducedMatrixRow^^.PivotColumn
        then Break
      else
        begin
          if Vector2.Size<Vector.Count+ReducedMatrixRow^^.Row.Count then
            begin
              NewSize:=Vector2.Size;
              repeat
                NewSize:=2*NewSize
              until NewSize>=Vector.Count+ReducedMatrixRow^^.Row.Count;
              Vector2.Count:=0;
              ResizeSparseVector(Vector2,NewSize)
            end;
          Val:=@Vector.Components^[0].Val;
          PivotValue:=@ReducedMatrixRow^^.Row.Components^[0].Val;
          if eq_VV(Val^,PivotValue^) then
            begin
              CheckMax(MaxEl,@Vector,ReducedMatrixRow^^.Row.MaxEl);
              SparseSub(Vector,ReducedMatrixRow^^.Row,Vector2)
            end
          else if eq_VV_neg(Val^,PivotValue^) then
            begin
              CheckMax(MaxEl,@Vector,ReducedMatrixRow^^.Row.MaxEl);
              SparseAdd(Vector,ReducedMatrixRow^^.Row,Vector2)
            end
          else if Divisible_F(Factor,Val^,PivotValue^) then
            begin
              Negate(Factor,Factor);
              {$ifndef USE_GMP}
              CheckMax(MaxEl,@Vector,CheckMult(ReducedMatrixRow^^.Row.MaxEl,Abs(Factor)));
              {$endif}
              SparseAddScaled(Vector,ReducedMatrixRow^^.Row,Factor,Vector2)
            end
          else if Divisible_F(Factor,PivotValue^,Val^) then
            begin
              if Negative(Val^)
                then SparseNeg(Vector)
                else Negate(Factor,Factor);
              if eq_VI(Val^,1) then Dec(PivotN1Count);
              VectorT:=ReducedMatrixRow^^.Row;
              CopySparseVector(Vector,ReducedMatrixRow^^.Row);
              if Vector.Size<VectorT.Size
                then
                  begin
                    DisposeSparseVector(Vector,0);
                    Vector:=VectorT
                  end
                else
                  begin
                    Vector.Count:=VectorT.Count;
                    for j:=0 to Vector.Count-1 do
                      begin
                        V_Done(Vector.Components^[j].Val);
                        Vector.Components^[j]:=VectorT.Components^[j]
                      end;
                    DisposeSparseVector(VectorT,Vector.Count)
                  end;
              {$ifndef USE_GMP}
              GetMax(Vector);
              SetMax(MaxEl,Vector.MaxEl);
              CheckMax(MaxEl,nil,CheckMult(ReducedMatrixRow^^.Row.MaxEl,Abs(Factor)));
              {$endif}
              SparseAddScaled(Vector,ReducedMatrixRow^^.Row,Factor,Vector2)
            end
          else
            begin
              ggTExt(V_Temp,PivotValue^,Val^,pg,vg,p0,v0);
              if eq_VI(V_Temp,1) then Dec(PivotN1Count);
              {$ifndef USE_GMP}
              GetMax(Vector);
              SetMax(MaxEl,CheckMult(Vector.MaxEl,Abs(vg)));
              CheckMax(MaxEl,nil,CheckMult(ReducedMatrixRow^^.Row.MaxEl,Abs(pg)));
              {$endif}
              SparseAddScaledBoth(ReducedMatrixRow^^.Row,Vector,pg,vg,Vector2);
              CopySparseVector(Vector2,VectorT);
              {$ifndef USE_GMP}
              SetMax(MaxEl,CheckMult(Vector.MaxEl,Abs(v0)));
              CheckMax(MaxEl,nil,CheckMult(ReducedMatrixRow^^.Row.MaxEl,Abs(p0)));
              {$endif}
              SparseAddScaledBoth(ReducedMatrixRow^^.Row,Vector,p0,v0,Vector2);
              DisposeSparseVector(ReducedMatrixRow^^.Row,0);
              ReducedMatrixRow^^.Row:=VectorT
            end;
          VectorT:=Vector;
          Vector:=Vector2;
          Vector2:=VectorT;
          if Vector.Count=0 then Break;
          if Index1>=Vector.Components^[0].Index then Abort ('Internal error in DoOneRow.');
          Index1:=Vector.Components^[0].Index;
          ReducedMatrixRow:=@ReducedMatrixRow^^.Next
        end;
    if Vector.Count<>0 then
      begin
        GetMax(Vector);
        Val:=@Vector.Components^[0].Val;
        if Negative(Val^) then SparseNeg(Vector);
        New(NewRow);
        if not eq_VI(Val^,1) then
          if Defer
            then
              begin
                NewRow^.Next:=DeferredVectors;
                NewRow^.Row:=Vector;
                DeferredVectors:=NewRow;
                Exit
              end
            else Inc(PivotN1Count);
        with NewRow^ do
          begin
            Next:=ReducedMatrixRow^;
            CopySparseVector(Vector,Row);
            PivotColumn:=Index1
          end;
        ReducedMatrixRow^:=NewRow;
        Inc(PivotCount)
      end;
    DisposeSparseVector(Vector,0);
    Inc(RowDoneCount)
  end;

begin
  RowCount:=0;
  RowDoneCount:=0;
  WriteStr(Msg,'Finding rank of boundary map d',Dim);
  ActivityReport (Commentator, Msg, LvlNormal, InternalDescription);
  StartActivity (Commentator, 'elimination', 'Triangulating matrix by modified Gauss elimination', LvlNormal, InternalDescription);
  ReducedMatrix:=nil;
  DeferredVectors:=nil;
  Vector2.Size:=Size0;
  NewSparseVector(Vector2,0);
  PivotCount:=0;
  PivotN1Count:=0;
  ReducedMatrixStartRow:=@ReducedMatrix;
  LastCol:=0;
  V_Init(Factor);
  V_Init(pg);
  V_Init(vg);
  V_Init(p0);
  V_Init(v0);
  V_Init(V_Temp);
  while (PivotCount-PivotN1Count<MaxRank) and _(GetNextRow) (BigFaces,SmallFaces,Vector,Size0,RowCount, RowDoneCount,RowStatus) do
    DoOneRow(Vector,True);
  while DeferredVectors<>nil do
    begin
      Temp:=DeferredVectors;
      DeferredVectors:=DeferredVectors^.Next;
      DoOneRow(Temp^.Row,False);
      Dispose(Temp)
    end;
  RowStatus;
  V_Done(Factor);
  V_Done(pg);
  V_Done(vg);
  V_Done(p0);
  V_Done(v0);
  V_Done(V_Temp);
  DisposeSparseVector(Vector2,0);
  StopActivity (Commentator, 'Matrix triangulated', LvlNormal, InternalDescription);
  WriteStr(Msg,'Rank d',Dim,' = ',PivotCount);
  ActivityReport (Commentator, Msg, LvlImp, PartialResult);
  TorsionList:=nil;
  if (PivotN1Count=0) or (Mode = hmBetti)
    then DisposeMatrix(ReducedMatrix)
    else
      begin
        WriteStr (Msg, 'Searching for torsion of rank <= ',PivotN1Count);
        ActivityReport (Commentator, Msg, LvlImp, InternalDescription);
        StartActivity (Commentator, 'reduce', 'Reducing matrix', LvlNormal, InternalDescription);
        i:=0;
        TorsionMatrix:=nil;
        while ReducedMatrix<>nil do
          begin
            Temp:=ReducedMatrix;
            ReducedMatrix:=ReducedMatrix^.Next;
            if eq_VI(Temp^.Row.Components^[0].Val,1)
              then
                begin
                  DisposeSparseVector(Temp^.Row,0);
                  Dispose(Temp)
                end
              else
                begin
                  SparseReduce(Temp^.Row,ReducedMatrix);
                  Temp^.Next:=TorsionMatrix;
                  TorsionMatrix:=Temp;
                  Inc(i);
                  ProgressReport (Commentator, 'rows', LvlUnimp, i, PivotN1Count)
                end
          end;
        StopActivity (Commentator, 'Matrix reduced', LvlNormal, InternalDescription);
        GetTorsion(TorsionMatrix,PivotN1Count,@TorsionList,Mode,Coeff)
      end
end;

procedure _(WriteStrTorsion) (var s:String;TorsionList:PTorsionList;Flag:Boolean);
var Temp : TString;
begin
  s:='';
  while TorsionList<>nil do
    begin
      if Flag then s:=s+' + ';
      Flag:=True;
      s:=s+'Z_';
      V_WriteStr(Temp,TorsionList^.TVal);
      s:=s+Temp;
      if TorsionList^.Count<>1 then
        begin
          WriteStr(Temp,s,'^',TorsionList^.Count);
          s := Temp
        end;
      TorsionList:=TorsionList^.Next
    end
end;

procedure WriteTorsionList(var OutFile:Text;TorsionList:PTorsionList);
begin
  while TorsionList<>nil do
    begin
      Write(OutFile,', [');
      V_Write(OutFile,TorsionList^.TVal);
      Write(OutFile,',',TorsionList^.Count,']');
      TorsionList:=TorsionList^.Next
    end
end;

procedure _(WriteHomology) (var OutFile:Text;const Homology:_(THomTable); MinDim, MaxDim : TCount);
var j:Dim1;
begin
  Write(OutFile,'[');
  for j:=MinDim to MaxDim do
    begin
      if j<>MinDim then Write(OutFile,', ');
      with Homology[j] do
        begin
          Write(OutFile,'[',Betti);
          WriteTorsionList(OutFile,Torsion);
          Write(OutFile,']')
        end
    end;
  Write(OutFile,']')
end;

{$ifdef USE_GMP}
to begin do
  begin
    mpz_init(mpz_tmp);
    mpz_init(mpz_tmp2);
    V_init(V_Tmp1);
    V_init(V_Tmp2)
  end;

to end do
  begin
    mpz_clear(mpz_tmp);
    mpz_clear(mpz_tmp2);
    V_Done(V_Tmp1);
    V_Done(V_Tmp2)
  end;
{$endif}
end.
