// =================================================================== //
// Givaro : Prime numbers
//              Modular powering,
//              Fermat numbers,
//              Primality tests, Factorization one by one :
//                      (There are parameters to fix)
// Time-stamp: <21 Aug 99 14:37:50 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //
#ifndef _GIVARO_INTEGERS_PRIME_
#define _GIVARO_INTEGERS_PRIME_

#include <math.h>
#include "timer.h"
#include "integer.h"
#include "intprime.h"
#include "compgmp.h"


// =================================================================== //
// Modular powering
// =================================================================== //

const Integer powmod( const Integer& ti, const Integer l, const Integer& fn) {
    Integer res = Integer(one);
    Integer puiss = ti % fn;
    Integer p = ABS(l);
    Integer q,r;

    if (p>0) {
        while (p != 0) {
            divide(p,Integer(2),q,r);
            if (r == 1) 
                res = (res * puiss) % fn;
            p = q;
            puiss = (puiss * puiss) % fn;
        }
    }
    if (l < 0) {
	gcd(res,fn,q,r);
	return q;
    } else
    	return res ;
}

// =================================================================== //
// Fermat numbers
// =================================================================== //

const Integer fermat ( const long  n ) {
        // fermat(n) = 2^(2^n) + 1
    return pow(Integer(2), long(pow(double(2),double(n))) ) + 1;
}
   
int pepin (const long n) {
        // Fermat number primality test
    Integer fn;
    fn = fermat(n);
    Integer y;
    y = fn - powmod(3,(fn-1)/2,fn);
    return y == 1;
}


// =================================================================== //
// Random generator
// =================================================================== //

const Integer Random_Integer (Integer& r, const Integer& size) {
    Integer tmp = 1;
    long pas = 10007;
    Integer mpas = 1;
    while (mpas < size) {
        tmp += lrand48() * mpas;
        mpas *= pas;
    }
    return r = (tmp*tmp+(lrand48() * pas)) % size;
};

const Integer Random_Integer (Integer& r, long size) {
    Integer tmp = 1, mpas = 1;
    long pas = 10, npas = 0, a;
    while (npas++ < size) {
        a = lrand48();
        a = a-(a/pas)*pas;
        tmp += a * pas * mpas;
        mpas *= pas;
    }
    return r=tmp;
};

// =================================================================== //
// Primality tests and factorization algorithms
// =================================================================== //

// =================================================================== //
// Primality tests 
// =================================================================== //

int isprime_Miller(const Integer& n) {
        // Monte Carlo algorithm
        // returns 1    : n prime with probability 3/4
        // returns 0    : n composite
    if (n < 2) return 0;
    if (n <= 3) return 1;
    Integer r=0,t=n-1,a,q;
    //srand48( BaseTimer::seed() );
    Random_Integer(a,n);
    long s=0;
    while(r == 0) {
        t = t / 2;
        r = t % 2;
        s++;
    }
    q = powmod(a,t,n);
    if ( (q==1) || (q == (n-1))) return 1;
    for(;s>1;s--) {
        q = (q*q) % n;
        if (q == (n-1)) return 1;
    }
    return 0;
}

    
const Integer test_Lehmann(const Integer& n) {
        // Monte Carlo algorithm
        // returns n-1  : n prime with probability 1/2
        // returns 1    : n composite with probability 1/2
        // else         : n composite
    Integer A;
    //srand48(BaseTimer::seed());
    Random_Integer(A,n);
    return powmod(A,(n-1)/2,n);
}

int isprime_Lehmann(const Integer& n) {
    if (n < 2) return 0;
    if (n <= 3) return 1;
    Integer tmp = test_Lehmann(n);
    if (tmp == (n-1))
        return 1;
    return 0;
}



int isprime_Erathostene(const Integer& p) {
        // Deterministic algorithm
        // Valid for p < BOUNDARY_isprime
    int n = p%int(BOUNDARY_isprime);
    if ((! (n & 0x1)) || (! (n%3)) || (n<2)) return 0;
// first composite by prime > 3 : 25
    if (n <= 23) return 1;
    int * IP = new int[n+1];
    int i;
// int are initialized to zero !!!
//    for(i=3; i<=n; i++)
//        IP[i++] = 0;
    i=3;
    int j, ii;
    while ( (i<=sqrt(n)) && (!(IP[n])) ) {
        ii= i<<1;
        j = i + ii;
        while (j<=n) {
            IP[j] = 1L;
            j+=ii;
        }
        j = i+1;
        while (IP[++j]) { j++;}
        i = j;
    }
    return !(IP[n]);
}


const Integer nextprime(const Integer& p) {
	Integer n = p;
	if (!(p%2))
		n = n + 1;
	else
		n = n + 2;
	while (! isprime(n) )
		n = n + 2;
	return n;
}

const Integer prevprime(const Integer& p) {
        Integer n = p;
        if (!(p%2))
                n = n - 1;
	else
                n = n - 2;
        while (! isprime(n) )
                n = n - 2;
        return n;
}
const Integer Erathostene(const Integer& p) {
        // Deterministic algorithm
        // Valid for p < BOUNDARY_factor
    long n = p%long(BOUNDARY_factor);
    if (! (n & 0x1)) return Integer(2);
    if (! (n%3)) return Integer(3);
    if (n <= 23) return p;
        
    int * IP = new int[n+1];
    int i;
    for(i=3; i<=n; i++)
        IP[i++] = 0;
    i=3;
    int j, ii;
    while (i<=sqrt(n)){
        ii=i<<1;
        j = i+ii;
        while (j<=n) {
            IP[j] = 1;
            j+=ii;
        }
        j = i+1;
        while (IP[++j]) { j++;}
        if (!(IP[n]))
            i = j;
        else
            return Integer(i);

    }
    return p;
}

const Integer Pollard(const Integer& n) {
        // average number of iterations < 13/8*sqrt( Pi*n/2)
        // Sometimes the factor isn't prime -- TO EXPLICIT
//    if ( (n<3) || isprime(n) ) return n;
    if (n<3) return n;
    if ( isprime(n) ) return n;
    Integer g=1, m=0, x, y, p, A;
    //srand48(BaseTimer::seed());
    Random_Integer(A,n);
    y=A;
    p=1;
    while(g == 1) {
        if((m+=1) == p) {
            x=y;
            p*=2;
        }
        y=Pollard_fct(y,n);
        g=gcd(y-x,n);
    }
    return g;
}

const Integer Pollard_easy(const Integer& n, const unsigned long threshold) {
    if (n<3) return n;
    if ( isprime(n) ) return n;
    Integer g=1, m=0, x, y, p, A;
    //srand48(BaseTimer::seed());
    Random_Integer(A,n);
    y=A;
    p=1;
    unsigned long c = 0;
    while ( (g == 1) && (++c < threshold)) {
        if((m+=1) == p) {
            x=y;
            p*=2;
        }
        y=Pollard_fct(y,n);
        g=gcd(y-x,n);
    }
    if (g == 1)
	return n;
    else
	return g;
}

 
#endif
