// ======================================================================= //
// Copyright (C) Linbox project 1999
// Domain Wiedemann
// Multiply computations are stopped when the polynomials remains the same
// for more than EARLY_TERM_THRESOLD
// Time-stamp: <01 Aug 00 15:17:26 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //
#ifndef _LIN_DOM_WIEDEMANN_C_
#define _LIN_DOM_WIEDEMANN_C_

#include <poly1.h>
#include <givtype.h>
#include <givrandom.h>

#define EARLY_TERM_THRESHOLD 20

template<class BlackBoxDomain> class WiedemannDom ;

template<class BlackBoxDomain> class WiedemannDom {
public:
    typedef          BlackBoxDomain                     BlackBoxDomain_t;
    typedef typename BlackBoxDomain::Rep                Rep;
//     typedef typename Rep::PreferredInMatrix_t           PreferredInMatrix_t;
//     typedef typename Rep::PreferredOutMatrix_t          PreferredOutMatrix_t;
    typedef typename BlackBoxDomain::PreferredInMatrix_t           PreferredInMatrix_t;
    typedef typename BlackBoxDomain::PreferredOutMatrix_t          PreferredOutMatrix_t;
    typedef typename BlackBoxDomain::Domain_t           Domain;
    typedef typename BlackBoxDomain::Type_t             Type_t;
    typedef          WiedemannDom< BlackBoxDomain >     Self_t;
private:
        // Berlekamp/Massey's Domain can differ from the BlackBox Domain.
        // (really ?)
    Domain _domain;
    BlackBoxDomain _BB_domain;

public:
        //-- Constructors
    WiedemannDom() : _BB_domain(), _domain() {};
    WiedemannDom(const Self_t& M) : _BB_domain(M._BB_domain), _domain(M._domain) {}
    WiedemannDom(const BlackBoxDomain& D) : _BB_domain(D), _domain(D.getdomain()) {}  
    WiedemannDom(const BlackBoxDomain& MD, const Domain& D) : _BB_domain(MD), _domain(D) {}  

        //-- Principal method
    template<class Polynome>
    void operator() (Polynome& C, const Rep& A, PreferredInMatrix_t& u, bool i) {
        massey(C, A, u, i);
    };
    

private:

// -----------------------------------------------
// dot product
// Preconditions : u.size() > 0, v.size() > 0 ....
// -----------------------------------------------
    template <class A1, class A2>
    void DOTPROD(Type_t& coeff, const A1& u, const A2& v) {
        _domain.mul(coeff, u[0], v[0]);
        for (Indice k=u.size()-1 ; k>0; --k)
            _domain.axpyin(coeff, u[k], v[k]);
    }
    
// -------------------------------------------------------------------
// Berlekamp/Massey algorithm with Wiedemann's Sequence generation
// Preconditions : u.size() > 0, v.size() > 0 ....
// -------------------------------------------------------------------
    template<class Polynome>
    void massey(Polynome& C, const Rep& A, PreferredInMatrix_t& u, bool is_symetric) { 
        startActivity(GlobalCommentator, "massey", "Wiedemann algorithm", LVL_UNIMP, INTERNAL_DESCRIPTION ); 
//         const Indice ni = A.n_row(), nj = u.size();
        const Indice ni = A.size(), nj = u.size();
        const Indice n = GIVMIN(ni,nj);
        const Indice END = 2*n;

// ====================================================
// Sequence initialization
        PreferredOutMatrix_t S(END+1);
        Type_t dots;
        PreferredOutMatrix_t v(ni);

        Indice s = -1;

    
        DOTPROD(dots,u,u); S[0] = dots;
        s = 0;
        if (iszero(S[0])) {
            activityReport(GlobalCommentator, "Wiedemann's algorithm Failure : self-orthogonal vector, please try again.", LVL_IMP, INTERNAL_ERROR);
//             cerr << "Wiedemann's algorithm Failure : self-orthogonal vector, please try again." << endl;
            return;
        }
// ====================================================
    
        Indeter X("X");
        C = Polynome(X, Degree(0), _domain.one);
        Polynome B(X, Degree(0), _domain.one);

        Indice L=0;
        Type_t b = _domain.one, d, Ds;
        Degree x = 1, b_deg = 0, c_deg = 0, l_deg;

        for (Indice N=0; N<END; ++N) {
            d = S[N];
            for(Indice i=GIVMIN(L,c_deg);i;--i) _domain.axpyin(d,C[i],S[N-i]);
            if (iszero(d)) {
                ++x;
            } else {
                if (L > (N>>1)) {
// -----------------------------------------------
// C = C + (Polynome(X,x,-d/b) * B);
// 
                    _domain.divin( _domain.neg(Ds,d), b);
                    l_deg = GIVMAX(c_deg,x+b_deg);
                    if (l_deg > c_deg) {
                        Polynome TD(X,l_deg);
                        Indice i = l_deg;
                        if (x>c_deg) {
                            for(;i>=x;--i)
                                _domain.mul(TD[i], Ds, B[i-x]);
                            for(;i>c_deg;--i)
                                TD[i] = _domain.zero;
                        } else {
                            for(;i>c_deg;--i)
                                _domain.mul(TD[i], Ds, B[i-x]);
                            for(;i>=x;--i)
                                _domain.axpy(TD[i],Ds,B[i-x],C[i]) ;
                        }
                        for(;i>=0;--i)
                            TD[i] = C[i];
                        C = TD;
                    } else {
                        for(Indice i=b_deg;i>=0;--i)
                            _domain.axpyin(C[i+x],Ds,B[i]);
                    }
// -----------------------------------------------
                    c_deg = C.degree();
                    ++x;
                } else {
                    Polynome TC(X,c_deg,C[c_deg]);
                    for (Indice i=c_deg;i--;) TC[i] = C[i];
// -----------------------------------------------
// C = C + (Polynome(X,x,-d/b) * B);
// 
                    _domain.divin( _domain.neg(Ds,d), b);
                    l_deg = GIVMAX(c_deg,x+b_deg);
                    if (l_deg > c_deg) {
                        Polynome TD(X,l_deg);
                        Indice i = l_deg;
                        if (x>c_deg) {
                            for(;i>=x;--i)
                                _domain.mul(TD[i], Ds, B[i-x]);
                            for(;i>c_deg;--i)
                                TD[i] = _domain.zero;
                        } else {
                            for(;i>c_deg;--i)
                                _domain.mul(TD[i], Ds, B[i-x]);
                            for(;i>=x;--i)
                                _domain.axpy(TD[i],Ds,B[i-x],C[i]) ;
                        }
                        for(;i>=0;--i)
                            TD[i] = C[i];
                        C = TD; 
                    } else {
                        for(Indice i=b_deg;i>=0;--i)
                            _domain.axpyin(C[i+x],Ds,B[i]);
                    }
// -----------------------------------------------
                    L = N+1-L;
                    B = TC; 
                    b_deg = c_deg;
                    c_deg = C.degree();
                    b = d;
                    x = 1;
                }
            }
            if (x > EARLY_TERM_THRESHOLD) break; 
// ====================================================
// Next coefficient
// 
            if (is_symetric) {
                if (N & 0x1) {
                    u = v;
                    DOTPROD(dots,u,u); S[++s] = dots;
                } else {
                    _BB_domain.Apply(v,A,u);
                    DOTPROD(dots,v,u); S[++s] = dots;
                }
            } else {
                if (N & 0x1) {
                    _BB_domain.ApplyTrans( u, A, v); 
                    DOTPROD(dots,u,u); S[++s] = dots;
                } else {
                    _BB_domain.Apply(v, A, u);
                    DOTPROD(dots,v,v); S[++s] = dots;
                }
            }
// ====================================================
            if (! (N % 1000) )
                progressReport(GlobalCommentator, "matrix-vector products", LVL_NORMAL, N, END);
        }
        char * UT = new char[60]; sprintf(UT,"Rank %ld found", C.degree()-C.val());
        stopActivity(GlobalCommentator, UT, LVL_NORMAL, INTERNAL_DESCRIPTION ); 
    }


public:
// ---------------------------------------------
// Symetric Wiedemann

    void rank_ssw(ulong& rank, const Rep& A, PreferredInMatrix_t& u) {
        Poly1<Type_t> phi;
        massey(phi, A, u, 1);
        rank = phi.degree() - phi.val();
    }

    void seq_rank_ssw(ulong& rank, const Rep& A) {
//        srand48(BaseTimer::seed());
        PreferredInMatrix_t VY(A.n_col());
        for(Indice k=0; k<VY.size(); k++)
            _domain.GIVRANDOM(VY[k]);
        Poly1<Type_t> phi;
        massey(phi, A, VY, 1);
        rank = phi.degree() - phi.val();
    }

    void valence_ssw(Type_t& valence, ulong& rank, const Rep& A, PreferredInMatrix_t& u) {
        Poly1<Type_t>  phi;
        massey(phi, A, u, 1);
        valence = phi.leadcoef() / phi.tcoef() ;
        rank = phi.degree() - phi.val();
    }

    void minpoly_ssw(Poly1< Type_t > & phi, const Rep& A, PreferredInMatrix_t& u) {
        massey(phi, A, u, 1);
        if (! iszero(phi)) {
            Type_t trail = phi.tcoef();
            Degree dp = phi.degree();
            for(Degree i = dp >> 1;i > 0; --i) {
                phi[0] = phi[i] / trail;
                phi[i] = phi[dp-i] / trail;
                phi[dp-i] = phi[0];
            }
            phi[0] = phi[dp] / trail;
            phi[dp] = _domain.one;
        }
    }

    
// ---------------------------------------------
// Wiedemann


    void rank_sw(ulong& rank, const Rep& A, PreferredInMatrix_t& u) {
        Poly1<Type_t> phi;
        massey(phi, A, u, 0);
        rank = phi.degree() - phi.val();
    }

    void valence_sw(Type_t& valence, ulong& rank, const Rep& A, PreferredInMatrix_t& u) {
        Poly1<Type_t> phi;
        massey(phi, A, u, 0);
        valence = phi.leadcoef() / phi.tcoef() ;
        rank = phi.degree() - phi.val();
    }

    void minpoly_sw(Poly1< Type_t >& phi, const Rep& A, PreferredInMatrix_t& u) {
        massey(phi, A, u, 0);
        Type_t trail = phi.tcoef();
        Degree dp = phi.degree();
        for(Degree i = dp >> 1;i > 0; --i) {
            phi[0] = phi[i] / trail;
            phi[i] = phi[dp-i] / trail;
            phi[dp-i] = phi[0];
        }
        phi[0] = phi[dp] / trail;
        phi[dp] = Type_t(one);
    }

    void seq_rank_sw(UIndice& rank, const Rep& A) {
//        srand48(BaseTimer::seed());
        PreferredInMatrix_t VY(A.n_col());
        for(Indice k=0; k<VY.size(); k++)
            _domain.GIVRANDOM(VY[k]);
        rank_sw(rank, A, VY);
    };
 
    void seq_rank_sw(UIndice& rank, const Rep& A, unsigned long nj) {
//        srand48(BaseTimer::seed());
        PreferredInMatrix_t VY(nj);
        for(Indice k=0; k<VY.size(); k++)
            _domain.GIVRANDOM(VY[k]);
        rank_sw(rank, A, VY);
    };
 

// --------------------------------------------------
// Warning : Here A is not a BlackBox, A is modified
// --------------------------------------------------
// Effective computation of D1 * A * D2
// A's sparsity is preserved.
// --------------------------------------------------
    void precondition(Rep& A) {
        Indice ll=0;
//        srand48(BaseTimer::seed());
        PreferredOutMatrix_t diag_left(A.n_row());
        for(; ll<A.n_row(); ll++)
            _domain.NONZEROGIVRANDOM(diag_left[ll]);
 
        PreferredInMatrix_t diag_right(A.n_col());
        for(ll=0;ll<A.n_col(); ll++)
            _domain.NONZEROGIVRANDOM(diag_right[ll]);

        _BB_domain.rank_precondition(diag_left, A, diag_right);
    }

    void precondition(Rep& A, unsigned long ni, unsigned long nj) {
        Indice ll=0;
//        srand48(BaseTimer::seed());
        PreferredOutMatrix_t diag_left(ni);
        for(; ll<ni; ll++)
            _domain.NONZEROGIVRANDOM(diag_left[ll]);
 
        PreferredInMatrix_t diag_right(nj);
        for(ll=0;ll<nj; ll++)
            _domain.NONZEROGIVRANDOM(diag_right[ll]);

        _BB_domain.rank_precondition(diag_left, A, diag_right);
    }

};


    
#endif
