// =================================================================== //
// Givaro / Athapascan-1
// Irreducible polynomial finder
// Primitive root finder
// Time-stamp: <13 Jul 99 16:28:43 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //
#ifndef _GIVARO_POLY_PRIMITIVE_ROOT_
#define _GIVARO_POLY_PRIMITIVE_ROOT_

// Reasonible cyclotomic polynomial size
#define CYCLO_DEGREE_BOUND 1000
#define CYCLO_TIMES_FACTOR 8

#include <stdlib.h>
#include <list>
using std::list;
#include <integer.h>
#include <intnumtheo.h>
#include <poly1.h>
#include <poly1factor.C>
#include <poly1cyclo.C>

// ---------------------------------------------------------------
// Monic irreducible polynomial of degree n over Z/pZ
// having 2, 3 nonzero terms or dividing a cyclotomic polynomial
// of degree < CYCLO_DEGREE_BOUND or a random one.
// ---------------------------------------------------------------
template<class MT, class T>
Poly1<T>& random_irreducible( Poly1<T>& P, Degree n, MT MOD) {
        // P must provide an indeterminate
    T Zero(zero), un(one);
    Poly1<T> Res(P.indeter(),n,un);
    for(int i=n;--i;)
        Res[i] = Zero;

        // Search for an irreducible BINOMIAL : X^n + a
    for(MT a=1; a<MOD; ++a) {
        Res[0] = T(a);
        if (is_irreducible(MOD,Res))
            return P = Res;
    }

        // Search for an irreducible TRINOMIAL : X^n + b*X^i + a
    for(Degree d=2;d<n;++d) {
        for(MT b=1; b<MOD; ++b) {
            Res[d] = T(b);
            for(MT a=1; a<MOD; ++a) {
                Res[0] = T(a);
                if (is_irreducible(MOD,Res))
                    return P = Res;
            }
        }
        Res[d] = Zero;
    }
    
    //srand48(BaseTimer::seed());
        // Search for a monic irreducible Polynomial
        // with random elements
    do {
        RANDOMPOLY1(Res, n, P.indeter());
        Res[n] = un;
        for(MT a=1; a<MOD; ++a) {
            Res[0] = T(a);
            if (is_irreducible(MOD,Res))
                return P = Res;
        }
    } while(1);
}

template<class MT, class T>
Poly1<T>& withcyclo_irreducible( Poly1<T>& P, Degree n, MT MOD) {
        // P must provide an indeterminate
    T Zero(zero), un(one);
    Poly1<T> Res(P.indeter(),n,un);
    for(int i=n;--i;)
        Res[i] = Zero;

    //srand48(BaseTimer::seed());

        // Search for a cyclotomic irreducible Polynomial
    Integer In(n),IMOD(MOD),r(2);
    for (;r<CYCLO_DEGREE_BOUND; r+=un) {
        if (order(IMOD,r) == In) break;
    }
    if (r<CYCLO_DEGREE_BOUND) {
        Poly1<T> cr(P.indeter());
        cyclotomic(cr, (ulong)(r%(ulong)CYCLO_DEGREE_BOUND));
            // cr is the product of irreducible
            // polynomials over Z/pZ of which degree is n.
        P = cr;
        Degree dp = P.degree();
        if (dp == n)
            return P;
        else if (MOD > 2) {
                // MOD 2 factorization is harder than other searches
            while (P.degree() > n)
                P = SplitFactor(IMOD, P, n);
            return P;
        }
    }
    
        // Search for an irreducible BINOMIAL : X^n + a
    for(MT a=1; a<MOD; ++a) {
        Res[0] = a;
        if (is_irreducible(MOD,Res))
            return P = Res;
    }

        // Search for an irreducible TRINOMIAL : X^n + b*X^i + a
    for(Degree d=2;d<n;++d) {
        for(MT b=1; b<MOD; ++b) {
            Res[d] = b;
            for(MT a=1; a<MOD; ++a) {
                Res[0] = a;
                if (is_irreducible(MOD,Res))
                    return P = Res;
            }
        }
        Res[d] = Zero;
    }
    
        // Search for an irreducible Polynomial
        // with random elements
    do {
        RANDOMPOLY1(Res, n, P.indeter());
        Res[n] = un;
        for(MT a=1; a<MOD; ++a) {
            Res[0] = a;
            if (is_irreducible(MOD,Res))
                return P = Res;
        }
    } while(1);
}

// ---------------------------------------------------------------
// Primitive Root over Z/pZ / F
// ---------------------------------------------------------------

template<class MT, class T>
int is_prim_root( const Poly1<T>& P, Degree n, MT MOD, const Poly1<T>& F) {
    int isproot = 0;
    Poly1<T> A = P % F;
    Poly1<T> Pun(P.indeter(),Degree(0),T(one));
    if ( (gcd(A,F)).degree() == 0) {
        MT q = (pow(Integer(MOD),n)).Int2long();
        Integer qp(q - 1);
        list<Integer> L;
        Integer_factorset(qp,L);
        list<Integer>::iterator li = L.begin();
        isproot = 1;
        for(;(li != L.end()) && isproot; ++li)
            isproot = ( powmod(A, qp / *li, F) != Pun);
    }
    return isproot;
}

template<class MT, class T>
Integer order( const Poly1<T>& P, Degree n, MT MOD, const Poly1<T>& F) {
    int isproot = 0;
   
    Poly1<T> A = P % F;
    Poly1<T> Pun(P.indeter(),Degree(0),T(one));
    if ( (gcd(A,F)).degree() == 0) {
        Integer qp(pow(Integer(MOD),n) - 1),g,gg;
        list<Integer> L;
        Integer_factorset(qp,L);
        L.sort();
        list<Integer>::iterator li = L.begin();
        isproot = 1;
        for(;(li != L.end()) && isproot; ++li)
            isproot = ( powmod(A, g = (qp / *li), F) != Pun);
        if (isproot)
            return qp;
        else {
            for(;li!=L.end();++li)
                while (iszero(g % *li) && (powmod(A, gg = (g / *li), F) == Pun))
                    g = gg;
            return g;
        }
    }
    return Integer(zero);
}

template<class MT, class T>
Poly1<T>& random_prim_root( Poly1<T>& P, Poly1<T>& R, Degree n, MT MOD) {
    T Zero(zero), un(one);
    //srand48(BaseTimer::seed());

        // Search for a cyclotomic primitive Polynomial
    Integer IMOD(MOD), r = pow(IMOD,n) - 1, In(n);
        // Theorems : - order of x in Z/(x^n-1)Z = n !
        //            - cyclo(a).degree() = phi(a)
    if ((r<CYCLO_DEGREE_BOUND) && (phi(r) < CYCLO_TIMES_FACTOR*n)) {
        Poly1<T> cr(P.indeter());
        cyclotomic(cr, (ulong)(r%(ulong)CYCLO_DEGREE_BOUND));
            // cr is the product of irreducible
            // polynomials over Z/pZ of which degree is n.
        P = cr;
        Degree dp = P.degree();
        if (dp == n) {
            Poly1<T> RR(P.indeter(),Degree(1),un);
            RR[0] = Zero;
            return R = RR;
        } else if (MOD > 2) {
                // MOD 2 SplitFactor factorization is harder than other searches
                // because of bad probabilities.
            while (P.degree() > n)
                P = SplitFactor(IMOD, P, n);
            Poly1<T> RR(P.indeter(),Degree(1),un);
            RR[0] = Zero;
            return R = RR;
        }
    }
    

    random_irreducible(P,n,MOD);

        // Search for a primitive BINOMIAL : X^i + a
    for(Degree di=1;di<n;++di) {
        Poly1<T> Res(P.indeter(),di,un);
        for(MT a=MOD; a--; ) {
            Res[0] = T(a);
            if (is_prim_root(Res,n,MOD,P))
                return R = Res;
        }
    }
        // Search for a primitive TRINOMIAL : X^i + b*X^j + a
    for(Degree di=2;di<n;++di) {
        Poly1<T> Res(P.indeter(),di,un);
        for(Degree dj=1;dj<di;++dj)
            for(MT b=MOD; b--;) {
                Res[dj] = T(b);
                for(MT a=MOD; a--;) {
                    Res[0] = T(a);
                    if (is_prim_root(Res,n,MOD,P))
                        return R = Res;
                }
            }
    }

        // Search for a primitive Polynomial
        // with random elements
    Poly1<T> Res;
    do {
        RANDOMPOLY1(Res, n-1, P.indeter());
        Res[n-1] = un;
        for(MT a=1; a<MOD; ++a) {
            Res[0] = T(a);
            if (is_prim_root(Res,n,MOD,P))
                return R = Res;
        }
    } while(1);
}


template<class MT, class T>
Poly1<T>& random_prim_root( Poly1<T>& P, Degree n, MT MOD, const Poly1<T>& F) {
    T Zero(zero), un(one);
    //srand48(BaseTimer::seed());

    Poly1<T> Res(P.indeter(),Degree(1),un);

        // Search for a primitive BINOMIAL : X + a
    for(MT a=0; a<MOD; ++a) {
        Res[0] = a;
        if (is_prim_root(Res,n,MOD,F))
            return P = Res;
    }

        // Search for a primitive Polynomial
        // with random elements
        // (has probability 1/MOD to be one)
    do {
        RANDOMPOLY1(Res, n, P.indeter());
        Res[n] = un;
        for(MT a=0; a<MOD; ++a) {
            Res[0] = a;
            if (is_prim_root(Res,n,MOD,F))
                return P = Res;
        }
    } while(1);
}

#endif
