#ifndef _GIVARO_GFQ1_H_
#define _GIVARO_GFQ1_H_

// ==========================================================================
// file: givdomgfq.h 
// Time-stamp: <23 Aug 99 16:58:32 Jean-Guillaume.Dumas@imag.fr>
// (c) Givaro Team
// date: 1999
// version: 
// Description:
//   Arithmetic on GF(p^k), with p a prime number less than 2^15
// ==========================================================================

#include "givtype.h"
#include "integer.h"
#include <iostream>


  // ------------------------------------------------- class GFqDom

template<class TT> class GFqDom {
private:
    typedef typename Signed_Trait<TT>::unsigned_type UTT;
public:
  typedef TT Rep;
  typedef UTT Residu_t;

  // ----- Representation of vector of the element
  typedef Rep* Array;
  typedef const Rep* constArray;
  // ----- Constantes 
  Rep zero;
  Rep one;

GFqDom() ;
GFqDom( const UTT ) ;
GFqDom( const UTT , const UTT ) ;
~GFqDom();

  // Access to the modulus, caracteristic, size, exponent
UTT residu() const;
UTT caracteristic() const;
UTT exponent() const;
UTT size() const;

  // Initialization of elements
void init( Rep& a ) const;
  // Access to the element
Rep access( const Rep a ) const;
UTT write( const Rep a ) const;

  // Test operators
inline int operator== (const GFqDom<TT>& a) const ;
inline int operator!= (const GFqDom<TT>& a) const ;

  // Miscellaneous functions
  short iszero( Rep ) const;
  short isone ( Rep ) const;
  size_t length ( Rep ) const;

  // ----- Operations with reduction: r <- a op b mod p, r <- op a mod p
  Rep& mul (Rep& r, const Rep a, const Rep b) const;
  Rep& div (Rep& r, const Rep a, const Rep b) const;
  Rep& add (Rep& r, const Rep a, const Rep b) const;
  Rep& sub (Rep& r, const Rep a, const Rep b) const;
  Rep& neg (Rep& r, const Rep a) const;
  Rep& inv (Rep& r, const Rep a) const;

  Rep& mulin (Rep& r, const Rep a) const;
  Rep& divin (Rep& r, const Rep a) const;
  Rep& addin (Rep& r, const Rep a) const;
  Rep& subin (Rep& r, const Rep a) const;

  // ----- Operations with reduction: r <- a op b mod p, r <- op a mod p
  void mul (const size_t sz, Array r, constArray a, constArray b) const;
  void mul (const size_t sz, Array r, constArray a, Rep b) const;

  void div (const size_t sz, Array r, constArray a, constArray b) const;
  void div (const size_t sz, Array r, constArray a, Rep b) const;

  void add (const size_t sz, Array r, constArray a, constArray b) const;
  void add (const size_t sz, Array r, constArray a, Rep b) const;

  void sub (const size_t sz, Array r, constArray a, constArray b) const;
  void sub (const size_t sz, Array r, constArray a, Rep b) const;
  void neg (const size_t sz, Array r, constArray a) const;
  void inv (const size_t sz, Array r, constArray a) const;

  // -- axpy: r <- a * x + y mod p
  Rep& axpy (Rep& r, const Rep a, const Rep b, const Rep c) const;
  void axpy (const size_t sz, Array r, Rep a, constArray x, constArray y) const;
  void axpy (const size_t sz, Array r, Rep a, constArray x, Rep c) const;

  // -- axmy: r <- a * x - y mod p
  Rep& axmy (Rep& r, const Rep a, const Rep b, const Rep c) const;
  void axmy (const size_t sz, Array r, Rep a, constArray x, constArray y) const;
  void axmy (const size_t sz, Array r, Rep a, constArray x, Rep c) const;

  // -- axpyin: r <- r + a * x mod p
  Rep& axpyin (Rep& r, const Rep a, const Rep b) const;
  void axpyin (const size_t sz, Array r, Rep a, constArray x) const;

  // -- Misc: r <- a mod p
  void assign ( const size_t sz, Array r, constArray a ) const;
  Rep& assign(Rep& r, const Rep a) const;
  Rep& assign(Rep& r, const Integer a) const;

  // <- \sum_i a[i], return 1 if a.size() ==0,
  void reduceadd ( Rep& r, const size_t sz, constArray a ) const; 

  // <- \prod_i a[i], return 1 if a.size() ==0,
  void reducemul ( Rep& r, const size_t sz, constArray a ) const; 

  // <- \sum_i a[i] * b[i] 
  void dotprod ( Rep& r, const size_t sz, constArray a, constArray b ) const; 
//   void dotprod ( Rep& r, const int bound, const size_t sz, constArray a, constArray b ) const; 

//   // ----- a -> r: uint16 to double
//   void i2d ( const size_t sz, double* r, constArray a ) const; 

//   // ----- a -> r % p: double to uint16 % p
//   void d2i ( const size_t sz, Array r, const double* a ) const; 

  // --- IO methods
  std::istream& read ( std::istream& s );
  std::ostream& write( std::ostream& s ) const;
  std::istream& read ( std::istream& s, Rep& a ) const;
  std::ostream& write( std::ostream& s, const Rep a ) const;

   // ----- random generators
    Rep& NONZEROGIVRANDOM(Rep&) ;
    Rep& GIVRANDOM(Rep&) ;
    

protected:
    UTT _caracteristic ;
    UTT _exponent ;
    UTT _q ;
    UTT _qm1 ;
    UTT _qm1o2;

        // G is a generator of GF(q)
        // p is GF(q)'s caracteristic
        // log2pol[ i ] = G^i(p)
        // pol2log[ j ] = i such that log2pol[i] = j
        // plus1[i] = k such that G^i + 1 = G^k
//    UTT* _log2pol;
//    UTT* _pol2log;
//    TT* _plus1;
	vector<UTT> _log2pol;
	vector<UTT> _pol2log;
	vector<UTT> _plus1;


  static void Init();
  static void End();
};


#include "givdomgfq.inl"

#endif
