#ifndef _ZPZ1_H_
#define _ZPZ1_H_
// ==========================================================================
// file: zpz.h 
// author: T. Gautier
// (c) PAC Team
// date: 
// version: 
// Description:
//   Arithmetic on Z/pZ, with p a prime number less than 2^15
//   Modulo typedef is a signed long number. In case it was modified
//   then bezout algorithm must be changed (coefficient can be negative).
// Bugs:
// ==========================================================================

#include "givtype.h"
#include <iostream>


  // Type for any element of all Z/pZ classes (i.e Zpz) 
  // for any p. Assume that any p are less than 2^15
typedef long Modulo ;


  // ------------------------------------------------- class Zpz 
  // An element of Zpz is a couple (r,p) of type Modulo.
  // This values are :
  // - r : residu.
  // - p : prime number within the computations are done.
  //   if p ==0 then r is any integer of long type in C++.

class Zpz {
public:
inline Zpz (const Neutral n=zero) : r(n) {};
  Zpz (const Modulo val) ;
  Zpz (const Modulo val, const Modulo P) ;

inline Zpz(const Zpz& a)  : r(a.r) {};

inline operator const Modulo() const ; 

  // defaults = between Zpz's 
  //Zpz& operator= (const Zpz& a) ;

  // Recopy operators
  Zpz& logcpy(const Zpz& a) ;
  Zpz& memcpy(const Zpz& a) ;

  // Return of the residu
inline Modulo residu() const { return r ; }

  // Return of the primal number
static void setZpz( const Modulo P ) ;
static void getZpz( Modulo& P ) { P = Zpz::p ; }

  // Assignment operators
inline Zpz& operator= (const int i) ;
inline Zpz& operator= (const long i) ;

  // Test operators
inline int operator== (const Zpz& a) const ;
inline int operator!= (const Zpz& a) const ;

  // Arithmetic operator:
inline const Zpz operator* (const Zpz& ) const ;
inline const Zpz operator/ (const Zpz& ) const ;
inline const Zpz operator+ (const Zpz& ) const ;
inline const Zpz operator- (const Zpz& ) const ;
inline const Zpz operator- () const ;


inline Zpz& operator*= (const Zpz& ) ;
inline Zpz& operator/= (const Zpz& ) ;
inline Zpz& operator+= (const Zpz& ) ;
inline Zpz& operator-= (const Zpz& ) ;
inline Zpz& muladd (const Zpz&, const Zpz&, const Zpz&) ;
inline Zpz& automuladd (const Zpz&, const Zpz&) ;

  // Miscellaneous functions
friend inline int   iszero(const Zpz& a)  ;
friend inline int   isone (const Zpz& a)  ;
friend inline ulong length(const Zpz& a)  ;

friend inline std::istream& operator>> (std::istream& i, Zpz& a) ;
friend inline std::ostream& operator<< (std::ostream& o, const Zpz a) ;

  // Implementation functions : between array and element
inline static void mul (Zpz& r,  Zpz a,  Zpz b);
       static void mul (int size, Zpz* r,  const Zpz* a,  const Zpz* b);
       static void mul (int size, Zpz* r,  const Zpz* a,  const Zpz b);

inline static void div (Zpz& r,  Zpz a,  Zpz b);
       static void div (int size, Zpz* r,  const Zpz* a,  const Zpz* b);
       static void div (int size, Zpz* r,  const Zpz* a,  const Zpz b);

inline static void add (Zpz& r,  Zpz a,  Zpz b);
       static void add (int size, Zpz* r,  const Zpz* a,  const Zpz* b);
       static void add (int size, Zpz* r,  const Zpz* a,  const Zpz b);

inline static void sub (Zpz& r,  Zpz a,  Zpz b);
       static void sub (int size, Zpz* r,  const Zpz* a,  const Zpz* b);
       static void sub (int size, Zpz* r,  const Zpz* a,  const Zpz b);

inline static void neg (Zpz& r,  Zpz a);
       static void neg (int size, Zpz*  r,  const Zpz* a);

       static void inv (Modulo& r,  const Modulo a, const Modulo p);
       static void inv (Zpz& r, const Zpz& a, const Modulo p);
       static void inv (Zpz& r, const Zpz& a );

static Modulo bezout (Modulo a, Modulo b, Modulo& u, Modulo& v) ;

protected:
    Modulo r ;   // residu  = m % p
    
public:
    static Modulo p ;
};

#include "zpz.inl"

#endif
