// ==========================================================================
// file: zpz.inl
// author: T. Gautier
// (c) GIVARO Team
// date: 
// version: 
// Description:
//   Arithmetic on Z/pZ, with p a prime number less than 2^15
//   Modulo typedef is a signed long number. In case it was modified
//   then bezout algorithm must be changed (coefficient can be negative).
// Bugs:
// ==========================================================================

  // In the next functions r,a,b are of type Modulo or Zpz
  // Arithmetic operators : compute r = a <op> b mod p.
  // Assume that 0 < p, 0 < r, 0<a, 0<b and returns a result
  // in the interval [0,p-1]
#define MODULO_ADD(res, a, b, p) { (res) = (a) + (b) -  (p) ; \
                                   (res) = (res) < 0 ? (res) + (p) : (res) ;}
#define MODULO_SUB(res, a, b, p) { (res) = (a) - (b) ; \
                                   (res) = (res) >= 0 ? (res) : (res) + p ; }
#define MODULO_MUL(res, a, b, p) { (res) = ((ulong)((a) * (b))) % (ulong)p ;}
#define MODULO_DIV(res, a, b, p) { MODULO_INV(res, b, p) ; MODULO_MUL(res,a,res,p) ; }
#define MODULO_INV(res, a, p)    {   Zpz::inv((res), (a), (p)); }
#define MODULO_NEG(res, a, p)    { (res) = ((a) ==0 ? 0 : p - (a)) ; }
#define MODULO_MULADD(res, a, b, c, p)    { (res) = (ulong)((a) * (b) + (c)) % (ulong)p; }

#define ZPZ_ADD(res, a, b, p) { (res).r = (a).r + (b).r -  (p) ; \
                                (res).r = (res).r < 0 ? (res).r + (p) : (res).r ;}
/*
#define ZPZ_ADD(res, a, b, p) { (res).r = (a).r + (b).r ; \
                                (res).r = (res).r < p ? (res).r : (res).r - p ; }
*/
#define ZPZ_SUB(res, a, b, p) { (res).r = (a).r - (b).r ; \
                                (res).r = (res).r >= 0 ? (res).r : (res).r + p ; }
#define ZPZ_MUL(res, a, b, p) { (res).r = ((ulong)((a).r * (b).r)) % (ulong)p ;}
#define ZPZ_DIV(res, a, b, p) {  MODULO_INV((res).r, (b).r, (p)) ; ZPZ_MUL((res),(a),(res),(p)) ; }
#define ZPZ_NEG(res, a, p)    { (res).r = ((a).r ==0 ? 0 : p - (a).r) ; }
#define ZPZ_MULADD(res, a, b, c, p) {MODULO_MULADD((res).r,(a).r,(b).r,(c).r,p);}

  // Recopy operators
inline Zpz& Zpz::logcpy(const Zpz& a) 
  { return *this = a; }

inline Zpz& Zpz::memcpy(const Zpz& a) 
  { return *this = a; }

inline Zpz& Zpz::operator= (const int i)
  { *this = Zpz(i) ; return *this ; }

inline Zpz& Zpz::operator= (const long i)
  { *this = Zpz(i) ; return *this ; }

inline int Zpz::operator== (const Zpz& a) const 
  { return (r == a.r) ; }

inline int Zpz::operator!= (const Zpz& a) const 
  { return r != a.r ; }


 // Arithmetics operators : 
 // ------------ A*B + C --> R
inline Zpz& Zpz::muladd (const Zpz& a, const Zpz& b, const Zpz& c) 
  { ZPZ_MULADD(*this, a, b, c, Zpz::p) ; return *this ; }

inline Zpz& Zpz::automuladd (const Zpz& a, const Zpz& b) {
    ZPZ_MULADD(*this, a, b, *this, Zpz::p); return *this ; }

 // ------------ A * B -> R 
inline const Zpz Zpz::operator* (const Zpz& a) const
  { Zpz res ; ZPZ_MUL(res, *this, a, Zpz::p) ; return res ; }

inline Zpz& Zpz::operator*= (const Zpz& a) 
  { ZPZ_MUL(*this, *this, a, Zpz::p) ; return *this ; }

 // ------------ A + B -> R 
inline const Zpz Zpz::operator+ (const Zpz& a) const
  { Zpz res ; ZPZ_ADD(res, *this, a, Zpz::p) ; return res ; }

inline Zpz& Zpz::operator+= (const Zpz& a) 
  { ZPZ_ADD(*this, *this, a, Zpz::p) ; return *this ; }

 // ------------ A - B -> R 
inline const Zpz Zpz::operator- (const Zpz& a) const
  { Zpz res ; ZPZ_SUB(res, *this, a, Zpz::p) ; return res ; }

inline Zpz& Zpz::operator-= (const Zpz& a )
  { ZPZ_SUB(*this, *this, a, Zpz::p) ; return *this ; }

inline const Zpz Zpz::operator- () const
  { Zpz res ; ZPZ_NEG(res, *this, Zpz::p) ; return res ; }

 // ------------ A / B -> R 
inline const Zpz Zpz::operator/ (const Zpz& a ) const 
  { Zpz res ;
    ZPZ_DIV(res, *this, a, Zpz::p) ;
    return res ; }

inline Zpz& Zpz::operator/= (const Zpz& a )
  { ZPZ_DIV(*this, *this, a, Zpz::p) ; return *this ; }

 // ------------------------- Miscellaneous functions

inline Zpz::operator const Modulo() const
{ return r ; }

//inline Zpz::operator Modulo&() 
//{ return r ; }

inline  int iszero(const Zpz& a)
  { return a.r ==0 ; }

inline int isone(const Zpz& a)
  { return a.r ==1 ; }

inline ulong length(const Zpz& )
  { return sizeof(Modulo) ;}

inline std::istream& operator>> (std::istream& i, Zpz& a) 
  { Modulo l ; i >> l ; a = Zpz(l) ; return i ; }

inline std::ostream& operator<< (std::ostream& o, const Zpz a)
{
//  if (a.r >=0) o << '+' ; // JGD 23.02.98
  return o << a.r ;
}

  // ----------- Usefull method :
inline void Zpz::mul (Zpz& r, Zpz a, Zpz b)
  { ZPZ_MUL(r,a,b, Zpz::p) ; }

inline void Zpz::div (Zpz& r, Zpz a, Zpz b)
  { ZPZ_DIV(r, a, b, Zpz::p) ; }

inline void Zpz::add (Zpz& r,  Zpz a,  Zpz b)
  { ZPZ_ADD(r, a, b, Zpz::p) ; }

inline void Zpz::sub (Zpz& r, Zpz a,  Zpz b)
  { ZPZ_SUB(r, a, b, Zpz::p) ; }

inline void Zpz::neg (Zpz& r,  Zpz a)
  { ZPZ_NEG(r, a, Zpz::p) ; }

